from django.db import models

# This method has moved, and this is put here temporarily to make branch management easier
from ansible_base.rbac.management import create_dab_permissions as create_custom_permissions  # noqa


def give_permissions(apps, rd, users=(), teams=(), object_id=None, content_type_id=None):
    """
    Give user permission to an object, but for use in migrations
    rd - role definition to grant the user
    users - list of users to give this permission to
    teams - list of teams to give this permission to, can be objects or id list

    target object is implicitly specified by
    object_id - primary key of the object permission will apply to
    content_type_id - primary key of the content type for the object
    """
    ObjectRole = apps.get_model('dab_rbac', 'ObjectRole')

    # Create the object role and add users to it
    object_role_fields = dict(role_definition=rd, object_id=object_id, content_type_id=content_type_id)
    object_role, _ = ObjectRole.objects.get_or_create(**object_role_fields)

    if users:
        # Django seems to not process through_fields correctly in migrations
        # so it will use created_by as the target field name, which is incorrect, should be user
        # basically can not use object_role.users.add(actor)
        RoleUserAssignment = apps.get_model('dab_rbac', 'RoleUserAssignment')
        user_assignments = [
            RoleUserAssignment(object_role=object_role, user=user, **object_role_fields)
            for user in users
        ]
        RoleUserAssignment.objects.bulk_create(user_assignments, ignore_conflicts=True)
    if teams:
        RoleTeamAssignment = apps.get_model('dab_rbac', 'RoleTeamAssignment')
        # AWX has trouble getting the team object, conditionally accept team id list
        if isinstance(teams[0], models.Model):
            team_assignments = [
                RoleTeamAssignment(object_role=object_role, team=team, **object_role_fields)
                for team in teams
            ]
        else:
            team_assignments = [
                RoleTeamAssignment(object_role=object_role, team_id=team_id, **object_role_fields)
                for team_id in teams
            ]
        RoleTeamAssignment.objects.bulk_create(team_assignments, ignore_conflicts=True)



import importlib.util
import logging

from django.conf import settings

logger = logging.getLogger('ansible_base.lib.dynamic_config.dynamic_urls')


url_types = ['api_version_urls', 'root_urls', 'api_urls']
for url_type in url_types:
    globals()[url_type] = []

installed_apps = getattr(settings, 'INSTALLED_APPS', [])
for app in installed_apps:
    if app.startswith('ansible_base.'):
        if not importlib.util.find_spec(f'{app}.urls'):
            logger.debug(f'Module {app} does not specify urls.py')
            continue
        url_module = __import__(f'{app}.urls', fromlist=url_types)
        logger.debug(f'Including URLS from {app}.urls')
        for url_type in ['api_version_urls', 'root_urls', 'api_urls']:
            urls = getattr(url_module, url_type, [])
            globals()[url_type].extend(urls)



from flags.state import flag_state
from rest_framework import serializers

from .utils import get_django_flags


class FeatureFlagSerializer(serializers.Serializer):
    """Serialize list of feature flags"""

    def to_representation(self) -> dict:
        return_data = {}
        feature_flags = get_django_flags()
        for feature_flag in feature_flags:
            return_data[feature_flag] = flag_state(feature_flag)

        return return_data



import logging

from django.apps import apps as global_apps
from django.contrib.contenttypes.management import create_contenttypes
from django.db import DEFAULT_DB_ALIAS, router

from ansible_base.rbac import permission_registry

logger = logging.getLogger(__name__)


def create_dab_permissions(app_config, verbosity=2, interactive=True, using=DEFAULT_DB_ALIAS, apps=global_apps, **kwargs):
    """
    This is modified from the django auth.
    This will create DABPermission entries
    this will only create permissions for registered models
    """
    if not getattr(app_config, 'models_module', None):
        return

    # exit early if nothing is registered for this app
    app_label = app_config.label
    if not any(cls._meta.app_label == app_label for cls in permission_registry._registry):
        return

    # Ensure that contenttypes are created for this app. Needed if
    # 'ansible_base.rbac' is in INSTALLED_APPS before
    # 'django.contrib.contenttypes'.
    create_contenttypes(
        app_config,
        verbosity=verbosity,
        interactive=interactive,
        using=using,
        apps=apps,
        **kwargs,
    )

    try:
        app_config = apps.get_app_config(app_label)
        ContentType = apps.get_model("contenttypes", "ContentType")
        Permission = apps.get_model("dab_rbac", "DABPermission")
    except LookupError:
        return

    if not router.allow_migrate_model(using, Permission):
        return

    # This will hold the permissions we're looking for as (content_type, (codename, name))
    searched_perms = []
    # The codenames and ctypes that should exist.
    ctypes = set()
    for klass in app_config.get_models():
        if not permission_registry.is_registered(klass):
            continue
        # Force looking up the content types in the current database
        # before creating foreign keys to them.
        ctype = ContentType.objects.db_manager(using).get_for_model(klass, for_concrete_model=False)

        ctypes.add(ctype)

        for action in klass._meta.default_permissions:
            searched_perms.append(
                (
                    ctype,
                    (
                        f"{action}_{klass._meta.model_name}",
                        f"Can {action} {klass._meta.verbose_name_raw}",
                    ),
                )
            )
        for codename, name in klass._meta.permissions:
            searched_perms.append((ctype, (codename, name)))

    # Find all the Permissions that have a content_type for a model we're
    # looking for.  We don't need to check for codenames since we already have
    # a list of the ones we're going to create.
    all_perms = set(Permission.objects.using(using).filter(content_type__in=ctypes).values_list("content_type", "codename"))

    perms = []
    for ct, (codename, name) in searched_perms:
        if (ct.pk, codename) not in all_perms:
            permission = Permission()
            permission._state.db = using
            permission.codename = codename
            permission.name = name
            permission.content_type = ct
            perms.append(permission)

    Permission.objects.using(using).bulk_create(perms)
    for perm in perms:
        logger.debug("Adding permission '%s'" % perm)



TRIGGER_DEFINITION = {
    # If adding a top level change create_claims in common.py
    "always": {
        "type": {},
        "keys": {},
    },
    "never": {"type": {}, "keys": {}},
    "groups": {
        "type": {},
        "keys": {
            # If adding a group key be sure to modify process_groups in common.py
            "has_or": {
                "type": [],
                "contents": "",
            },
            "has_and": {
                "type": [],
                "contents": "",
            },
            "has_not": {
                "type": [],
                "contents": "",
            },
        },
    },
    "attributes": {
        "type": {},
        "keys": {
            # If adding a key or */keys be sure to modify process_user_attributes in common.py
            "join_condition": {
                "type": "",
                "choices": ["or", "and"],
            },
            "*": {
                "type": {},
                "keys": {
                    "contains": {
                        "type": "",
                    },
                    "matches": {
                        "type": "",
                    },
                    "ends_with": {
                        "type": "",
                    },
                    "equals": {
                        "type": "",
                    },
                    "in": {
                        "type": [],
                        "contents": "",
                    },
                },
            },
        },
    },
}



import logging
from datetime import datetime
from typing import Optional, Tuple

import jwt
from django.apps import apps
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model
from django.db.utils import IntegrityError
from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import AuthenticationFailed

from ansible_base.jwt_consumer.common.cache import JWTCache
from ansible_base.jwt_consumer.common.cert import JWTCert, JWTCertException
from ansible_base.jwt_consumer.common.exceptions import HTTP_498_INVALID_TOKEN, InvalidTokenException
from ansible_base.lib.logging.runtime import log_excess_runtime
from ansible_base.lib.utils.auth import get_user_by_ansible_id
from ansible_base.lib.utils.translations import translatableConditionally as _
from ansible_base.resource_registry.models import Resource, ResourceType
from ansible_base.resource_registry.signals.handlers import no_reverse_sync

logger = logging.getLogger("ansible_base.jwt_consumer.common.auth")


# These fields are used to both map the user as well as to validate the JWT token
default_mapped_user_fields = [
    "username",
    "first_name",
    "last_name",
    "email",
    "is_superuser",
]

_permission_registry = None


def permission_registry():
    global _permission_registry

    if not _permission_registry:
        from ansible_base.rbac.permission_registry import permission_registry as permission_registry_singleton

        _permission_registry = permission_registry_singleton
    return _permission_registry


class JWTCommonAuth:
    def __init__(self, user_fields=default_mapped_user_fields) -> None:
        self.mapped_user_fields = user_fields
        self.cache = JWTCache()
        self.user = None
        self.token = None

    @log_excess_runtime(logger, debug_cutoff=0.01)
    def parse_jwt_token(self, request):
        """
        parses the given request setting self.user and self.token
        """

        self.user = None
        self.token = None

        logger.debug("Starting JWT Authentication")
        if request is None:
            return

        token_from_header = request.headers.get("X-DAB-JW-TOKEN", None)
        request_id = request.headers.get("X-Request-Id")
        if not token_from_header:
            logger.debug("X-DAB-JW-TOKEN header not set for JWT authentication")
            return
        logger.debug(f"Received JWT auth token: {token_from_header}")

        cert_object = JWTCert()
        try:
            cert_object.get_decryption_key()
        except JWTCertException as jce:
            logger.error(jce)
            raise AuthenticationFailed(jce)

        if cert_object.key is None:
            return None, None

        try:
            self.token = self.validate_token(token_from_header, cert_object.key, request_id)
        except jwt.exceptions.DecodeError as de:
            # This exception means the decryption key failed... maybe it was because the cache is bad.
            if not cert_object.cached:
                # It wasn't cached anyway so we an just raise our exception
                self.log_and_raise(_("JWT decoding failed: %(e)s, check your key and generated token"), {"e": de})

            # We had a cached key so lets get the key again ignoring the cache
            old_key = cert_object.key
            try:
                cert_object.get_decryption_key(ignore_cache=True)
            except JWTCertException as jce:
                self.log_and_raise(_("Failed to get JWT token on the second try: %(e)s"), {"e": jce})
            if old_key == cert_object.key:
                # The new key matched the old key so don't even try and decrypt again, the key just doesn't match
                self.log_and_raise(_("JWT decoding failed: %(e)s, cached key was correct; check your key and generated token"), {"e": de})
            # Since we got a new key, lets go ahead and try to validate the token again.
            # If it fails this time we can just raise whatever
            self.token = self.validate_token(token_from_header, cert_object.key, request_id)

        # Let's see if we have the same user info in the cache already
        # Note: we're not using the "_, user_defaults=" trick here because _ is used for our translation function.
        user_defaults = self.cache.check_user_in_cache(self.token)[1]

        self.user = None
        try:
            self.user = get_user_by_ansible_id(self.token['sub'])
        except ObjectDoesNotExist:
            pass

        if not self.user:
            # Either the user wasn't cached or the requested user was not in the DB so we need to make a new one
            resource_kwargs = {}
            for resource_key, token_key in (('resource_data', 'user_data'), ('ansible_id', 'sub'), ('service_id', 'service_id')):
                if token_key not in self.token:
                    logger.warning(f'Missing {token_key} in JWT data, omitting {resource_key} from local resource entry')
                else:
                    resource_kwargs[resource_key] = self.token[token_key]
            try:
                resource = Resource.create_resource(ResourceType.objects.get(name="shared.user"), **resource_kwargs)
                self.user = resource.content_object
                logger.info(f"New user {self.user.username} created from JWT auth")
            except IntegrityError as exc:
                logger.debug(f'Existing user {self.token["user_data"]} is a conflict with local user, error: {exc}')
                with no_reverse_sync():
                    if user_defaults['is_superuser'] is False:
                        user_defaults.pop('is_superuser')
                    self.user, created = get_user_model().objects.update_or_create(
                        username=self.token["user_data"]['username'],
                        defaults=user_defaults,
                    )

                    resource = Resource.get_resource_for_object(self.user)

                    resource.ansible_id = self.token['sub']
                    resource.service_id = self.token['service_id']
                    resource.save(update_fields=['ansible_id', 'service_id'])

        setattr(self.user, "resource_api_actions", self.token.get("resource_api_actions", None))

        logger.info(f"User {self.user.username} authenticated from JWT auth")

    def log_and_raise(self, conditional_translate_object, expand_values={}, error_code=None):
        logger.error(conditional_translate_object.not_translated() % expand_values)
        translated_error_message = conditional_translate_object.translated() % expand_values
        if error_code == HTTP_498_INVALID_TOKEN:
            raise InvalidTokenException(translated_error_message)
        else:
            raise AuthenticationFailed(translated_error_message)

    def map_user_fields(self):
        if self.token is None or self.user is None:
            logger.error("Unable to map user fields because user or token is not defined, please call authenticate first")
            return

        user_needs_save = False
        for attribute in self.mapped_user_fields:
            old_value = getattr(self.user, attribute, None)
            new_value = self.token.get('user_data', {}).get(attribute, None)
            if old_value != new_value:
                if attribute == "username" and get_user_model().objects.filter(username=new_value).exists():
                    logger.warning(
                        f"Renaming user {old_value} to {new_value} would result in a duplicate key error. "
                        "Please make sure the sync task is running to prevent this warning in the future."
                    )
                    continue
                if attribute == 'is_superuser' and new_value is False:
                    continue
                logger.debug(f"Changing {attribute} for {self.user.username} from {old_value} to {new_value}")
                setattr(self.user, attribute, new_value)
                user_needs_save = True
        if user_needs_save:
            with no_reverse_sync():
                logger.info(f"Saving user {self.user.username}")
                self.user.save()

    def validate_token(self, unencrypted_token, decryption_key, request_id=None):
        validated_body = None

        # Decrypt the token
        try:
            logger.info("Decrypting token")
            validated_body = self.decode_jwt_token(
                unencrypted_token,
                decryption_key,
            )
        except jwt.exceptions.DecodeError as e:
            raise e  # This will be handled higher up
        except jwt.exceptions.ExpiredSignatureError:
            expired_token = self.decode_jwt_token(
                unencrypted_token,
                decryption_key,
                additional_options={"verify_exp": False},
            )
            expired_time = expired_token.get("exp")
            now = datetime.now().timestamp()
            time_diff = int(now - expired_time)
            self.log_and_raise(_(f"JWT expired {time_diff} seconds ago - check for clock skew. Request ID: {request_id}"), error_code=HTTP_498_INVALID_TOKEN)
        except jwt.exceptions.InvalidAudienceError:
            self.log_and_raise(_("JWT did not come for the correct audience"))
        except jwt.exceptions.InvalidIssuerError:
            self.log_and_raise(_("JWT did not come from the correct issuer"))
        except jwt.exceptions.MissingRequiredClaimError as e:
            self.log_and_raise(_("Failed to decrypt JWT: %(e)s"), {"e": e})
        except Exception as e:
            self.log_and_raise(_("Unknown error occurred decrypting JWT (%(e_class)s) %(e)s"), {"e_class": e.__class__, "e": e})

        logger.debug(validated_body)

        # Ensure all of the user pieces are part of the token
        missing_user_data = []
        for field in self.mapped_user_fields:
            if field not in validated_body['user_data']:
                missing_user_data.append(field)
        if missing_user_data:
            self.log_and_raise(_("JWT did not have proper user_data, missing fields: %(missing_fields)s"), {"missing_fields": ", ".join(missing_user_data)})

        # At this time we are not doing anything with regards to the version other than ensuring its there.

        return validated_body

    def decode_jwt_token(self, unencrypted_token, decryption_key, additional_options={}):
        local_required_field = ["sub", "user_data", "exp", "objects", "object_roles", "global_roles", "version"]
        options = {"require": local_required_field}
        options.update(additional_options)
        return jwt.decode(
            unencrypted_token,
            decryption_key,
            audience="ansible-services",
            options=options,
            issuer="ansible-issuer",
            algorithms=["RS256"],
        )

    def get_role_definition(self, name: str) -> Optional[Model]:
        """Simply get the RoleDefinition from the database if it exists and handler corner cases

        If this is the name of a managed role for which we have a corresponding definition in code,
        and that role can not be found in the database, it may be created here
        """
        from ansible_base.rbac.models import RoleDefinition

        try:
            return RoleDefinition.objects.get(name=name)
        except RoleDefinition.DoesNotExist:

            constructor = permission_registry().get_managed_role_constructor_by_name(name)
            if constructor:
                rd, _ = constructor.get_or_create(apps)
                return rd
        return None

    def process_rbac_permissions(self):
        """
        This is a default process_permissions which should be usable if you are using RBAC from DAB
        """
        if self.token is None or self.user is None:
            logger.error("Unable to process rbac permissions because user or token is not defined, please call authenticate first")
            return

        from ansible_base.rbac.models import RoleUserAssignment

        role_diff = RoleUserAssignment.objects.filter(user=self.user, role_definition__name__in=settings.ANSIBLE_BASE_JWT_MANAGED_ROLES)

        for system_role_name in self.token.get("global_roles", []):
            logger.debug(f"Processing system role {system_role_name} for {self.user.username}")
            rd = self.get_role_definition(system_role_name)
            if rd:
                if rd.name in settings.ANSIBLE_BASE_JWT_MANAGED_ROLES:
                    assignment = rd.give_global_permission(self.user)
                    role_diff = role_diff.exclude(pk=assignment.pk)
                    logger.info(f"Granted user {self.user.username} global role {system_role_name}")
                else:
                    logger.error(f"Unable to grant {self.user.username} system level role {system_role_name} because it is not a JWT managed role")
            else:
                logger.error(f"Unable to grant {self.user.username} system level role {system_role_name} because it does not exist")
                continue

        for object_role_name in self.token.get('object_roles', {}).keys():
            rd = self.get_role_definition(object_role_name)
            if rd is None:
                logger.error(f"Unable to grant {self.user.username} object role {object_role_name} because it does not exist")
                continue
            elif rd.name not in settings.ANSIBLE_BASE_JWT_MANAGED_ROLES:
                logger.error(f"Unable to grant {self.user.username} object role {object_role_name} because it is not a JWT managed role")
                continue

            object_type = self.token['object_roles'][object_role_name]['content_type']
            object_indexes = self.token['object_roles'][object_role_name]['objects']

            for index in object_indexes:
                object_data = self.token['objects'][object_type][index]
                try:
                    resource, obj = self.get_or_create_resource(object_type, object_data)
                except IntegrityError as e:
                    logger.warning(
                        f"Got integrity error ({e}) on {object_data}. Skipping {object_type} assignment. "
                        "Please make sure the sync task is running to prevent this warning in the future."
                    )
                    continue

                if resource is not None:
                    assignment = rd.give_permission(self.user, obj)
                    role_diff = role_diff.exclude(pk=assignment.pk)
                    logger.info(f"Granted user {self.user.username} role {object_role_name} to object {obj.name} with ansible_id {object_data['ansible_id']}")

        # Remove all permissions not authorized by the JWT
        for role_assignment in role_diff:
            rd = role_assignment.role_definition
            content_object = role_assignment.content_object
            if content_object:
                rd.remove_permission(self.user, content_object)
            else:
                rd.remove_global_permission(self.user)

    def get_or_create_resource(self, content_type: str, data: dict) -> Tuple[Optional[Resource], Optional[Model]]:
        """
        Gets or creates a resource from a content type and its default data

        This can only build or get organizations or teams
        """
        object_ansible_id = data['ansible_id']
        try:
            resource = Resource.objects.get(ansible_id=object_ansible_id)
            logger.debug(f"Resource {object_ansible_id} already exists")
            return resource, resource.content_object
        except Resource.DoesNotExist:
            pass

        # The resource was missing so we need to create its stub
        if content_type == 'team':
            # For a team we first have to make sure the org is there
            org_id = data['org']
            organization_data = self.token['objects']["organization"][org_id]

            # Now that we have the org we can build a team
            org_resource, _ = self.get_or_create_resource("organization", organization_data)

            resource = Resource.create_resource(
                ResourceType.objects.get(name="shared.team"),
                {"name": data["name"], "organization": org_resource.ansible_id},
                ansible_id=data["ansible_id"],
            )

            return resource, resource.content_object

        elif content_type == 'organization':
            resource = Resource.create_resource(
                ResourceType.objects.get(name="shared.organization"),
                {"name": data["name"]},
                ansible_id=data["ansible_id"],
            )

            return resource, resource.content_object
        else:
            logger.error(f"build_resource_stub does not know how to build an object of type {type}")
            return None, None


class JWTAuthentication(BaseAuthentication):
    map_fields = default_mapped_user_fields
    use_rbac_permissions = False

    def __init__(self):
        self.common_auth = JWTCommonAuth(self.map_fields)

    def authenticate(self, request):
        self.common_auth.parse_jwt_token(request)

        if self.common_auth.user:
            self.process_user_data()
            self.process_permissions()

            return self.common_auth.user, None
        else:
            return None

    def process_user_data(self):
        self.common_auth.map_user_fields()

    def process_permissions(self):
        if self.use_rbac_permissions:
            self.common_auth.process_rbac_permissions()
        else:
            logger.info("process_permissions was not overridden for JWTAuthentication")



import functools

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils.http import urlencode
from django.utils.translation import gettext_lazy as _

from ansible_base.lib.abstract_models import ImmutableCommonModel
from ansible_base.lib.utils.response import get_relative_url


class Entry(ImmutableCommonModel):
    """
    An activity stream entry.

    This is keyed on a generic object_id and content_type, which allows for
    a wide variety of objects to be used in the activity stream.
    """

    router_basename = 'activitystream'

    class Meta:
        verbose_name_plural = _('Entries')
        ordering = ['id']

    OPERATION_CHOICES = [
        ('create', _('Entity created')),
        ('update', _("Entity updated")),
        ('delete', _("Entity deleted")),
        ('associate', _("Entity was associated with another entity")),
        ('disassociate', _("Entity was disassociated with another entity")),
    ]

    content_type = models.ForeignKey(ContentType, on_delete=models.DO_NOTHING, help_text=_("The content type which was changed in this entry."))
    object_id = models.TextField(null=True, blank=True, help_text=_("The object id which was modified in this entry."))
    content_object = GenericForeignKey('content_type', 'object_id')
    operation = models.CharField(
        max_length=12, choices=OPERATION_CHOICES, help_text=_("The type of change recorded by the entry (i.e. create/update/delete/etc).")
    )
    changes = models.JSONField(null=True, blank=True, help_text=_("The changes to the system recorded by this entry."))

    # This is used for m2m (dis)associations
    related_content_type = models.ForeignKey(
        ContentType,
        on_delete=models.DO_NOTHING,
        null=True,
        blank=True,
        related_name='related_content_type',
        help_text=_("The content type if this entry is recording an many to many (M2M) change."),
    )
    related_object_id = models.TextField(
        null=True,
        blank=True,
        help_text=_("The object id of the related model if this entry is for an many to many (M2M) change."),
    )
    related_content_object = GenericForeignKey('related_content_type', 'related_object_id')
    related_field_name = models.CharField(
        max_length=64,
        null=True,
        blank=True,
        help_text=_("The related field name if this entry is for an many to many (M2M) relationship."),
    )

    def __str__(self):
        return f'[{self.created}] {self.get_operation_display()} by {self.created_by}: {self.content_type} {self.object_id}'

    @functools.cached_property
    def changed_fk_fields(self):
        """
        :return: A dictionary of {field_name: pk} for any ForeignKey fields that have changed.
            the pk is the new pk value, for changed fields. In the case of added/removed fields,
            the pk is the pk of the related object. In the case of a changed field, the pk is the
            new pk value.
        """
        changed_fks = {}

        if self.changes is None:
            return changed_fks

        for op in ('added_fields', 'changed_fields', 'removed_fields'):
            for field_name, value in self.changes.get(op, {}).items():
                if (model := self.content_type.model_class()) is None:
                    continue
                field = model._meta.get_field(field_name)
                if isinstance(field, models.ForeignKey):
                    try:
                        fk_model = field.related_model
                    except AttributeError:  # Likely the model was deleted
                        continue
                    # For added/removed fields, the value is the pk of the related object
                    # For changed fields, the value is the *new* pk value
                    pk = value[1] if op == 'changed_fields' else value
                    changed_fks[field_name] = (fk_model, pk)

        return changed_fks


class AuditableModel(models.Model):
    """
    A mixin class that provides integration to the activity stream from any
    model. A model should simply inherit from this class to have its
    create/update/delete events sent to the activity stream.
    """

    class Meta:
        abstract = True

    # Adding field names to this list will exclude them from the activity stream changes dictionaries
    activity_stream_excluded_field_names = []

    # Adding field names to this list will limit the activity stream changes dictionaries to only include these fields
    activity_stream_limit_field_names = []

    @property
    def activity_stream_entries(self):
        """
        A helper property that returns the activity stream entries for this object.
        """
        return Entry.objects.filter(content_type=ContentType.objects.get_for_model(self), object_id=self.pk).order_by('created')

    def extra_related_fields(self, request):
        content_type = ContentType.objects.get_for_model(self)
        query_kwargs = {
            'content_type': content_type.pk,
            'object_id': self.pk,
        }
        activity_stream_url = get_relative_url('activitystream-list') + '?' + urlencode(query_kwargs)
        return {
            'activity_stream': activity_stream_url,
        }



import logging
from typing import Optional

from django.conf import settings
from django.db.models import Field, ForeignKey, Model
from django.forms import model_to_dict
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import PermissionDenied

from ansible_base.lib.utils.db import ensure_transaction
from ansible_base.lib.utils.models import is_add_perm
from ansible_base.rbac.models import RoleDefinition
from ansible_base.rbac.permission_registry import permission_registry

logger = logging.getLogger(__name__)


def required_related_permission(field: Field) -> Optional[str]:
    """Permission level required to change field

    Returns a permission codename like change_inventory.
    """
    parent_field_name = permission_registry.get_parent_fd_name(field.model)
    if field.name == parent_field_name:
        return f'add_{field.model._meta.model_name}'
    rel_cls = field.related_model
    custom_perms = dict(rel_cls._meta.permissions)
    for action in settings.ANSIBLE_BASE_CHECK_RELATED_PERMISSIONS:
        codename = f'{action}_{rel_cls._meta.model_name}'
        if action in rel_cls._meta.default_permissions or codename in custom_perms:
            return codename
    return None


def related_permission_fields(cls):
    for field in cls._meta.concrete_fields:
        if isinstance(field, ForeignKey) and permission_registry.is_registered(field.related_model):
            yield field


def validate_field_data(field_name: str, data: dict):
    if f'{field_name}_id' in data:
        raise RuntimeError(f'Expected model_to_dict format of data, received {data}')
    if field_name in data and isinstance(data[field_name], Model):
        raise RuntimeError(f'Expected primary key for {field_name} but received {data[field_name]}')


def log_related_check(user, cls, errors, checked_fields, unchanged_fields):
    if errors:
        logger.warning(
            f'User {user.pk} lacks {cls._meta.model_name} related permissions, checked {checked_fields}, '
            f'errored: {list(errors.keys())}, unchanged: {unchanged_fields}'
        )
    elif checked_fields:
        logger.info(f'User {user.pk} has {cls._meta.model_name} related permissions {checked_fields}, unchanged: {unchanged_fields}')
    elif unchanged_fields:
        logger.debug(f'User {user.pk} needs no {cls._meta.model_name} related permissions, all fields unchanged: {unchanged_fields}')


def check_related_permissions(user, cls, old_data, new_data):
    """Raise PermissionDenied if user lacks access to changing related item

    Both old_data and new_data represent the properties of an object of cls.
    """
    errors = {}
    checked_fields = {}  # only for logging
    unchanged_fields = []  # only for logging

    for field in related_permission_fields(cls):
        # Assure that data structure is expected to avoid giving incorrect evaluations
        validate_field_data(field.name, new_data)
        validate_field_data(field.name, old_data)

        # A permission-relevant field is given in the new data
        if field.name in old_data and old_data.get(field.name) == new_data.get(field.name):
            unchanged_fields.append(field.name)
        else:
            # This field is verified to have changed compared to old data
            to_check = required_related_permission(field)
            if (to_check is None) or (field.null and (new_data.get(field.name) is None) and (not is_add_perm(to_check))):
                # user can null non-parent fields with no additional permission
                continue
            checked_fields[field.name] = to_check
            rel_obj = field.related_model(pk=new_data.get(field.name))
            if not user.has_obj_perm(rel_obj, to_check):
                errors[field.name] = _('You do not have permission to use this object.')

    # It is fairly useful to log the outcome for transparency to the administrator
    log_related_check(user, cls, errors, checked_fields, unchanged_fields)

    if errors:
        raise PermissionDenied(errors)


class RelatedAccessMixin:
    """Class to be used by apps to check permissions to related objects

    This is a core part of integrating permission checks with REST APIs.
    Full enablement of checks is accomplished by this plus
    the permission class.
    """

    def update(self, instance, validated_data):
        """Override DRF ModelSerializer.update method to check permissions

        The super() of this method does setattr on instance with validated_data
        thus, this is our last chance to get a representation of the prior
        object so we know what the user has changed and did not change.

        This runs in a transaction so that PermissionDenied exception will
        roll back changes.
        We have to save the model before we accurately know what the new
        fields are, because model logic can, and will, change things.
        """
        view = self.context.get('view', None)
        if not view or not view.request:
            logger.warning(f'Serializer cannot check related permissions of {self.Meta.model}-{instance.pk} because context was not passed to {type(self)}')
            updated_instance = super().update(instance, validated_data)
        else:
            # Properties of the prior instance must be saved before setattr with new data
            # this is analogous to making a copy of instance
            old_data = model_to_dict(instance)
            with ensure_transaction():
                updated_instance = super().update(instance, validated_data)
                check_related_permissions(view.request.user, self.Meta.model, old_data, model_to_dict(updated_instance))
        return updated_instance

    def create(self, validated_data):
        view = self.context.get('view', None)
        if not view or not view.request:
            logger.warning(f'Serializer cannot check related permissions for new {self.Meta.model} because context was not passed to {type(self)}')
            instance = super().create(validated_data)
        else:
            with ensure_transaction():
                instance = super().create(validated_data)
                check_related_permissions(view.request.user, self.Meta.model, {}, model_to_dict(instance))
                RoleDefinition.objects.give_creator_permissions(view.request.user, instance)
        return instance



# Generated by Django 4.2.8 on 2024-02-11 03:14

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_authentication', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticator',
            name='users',
            field=models.ManyToManyField(blank=True, help_text='The list of users who have authenticated from this authenticator', related_name='authenticators', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 4.2.16 on 2024-11-26 20:41

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_rbac', '0002_alter_objectrole_provides_teams_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dabpermission',
            name='codename',
            field=models.CharField(help_text='A codename for the permission, in the format {action}_{model_name}. Where action is typically the view set action (view/list/etc) from Django rest framework.', max_length=100, verbose_name='codename'),
        ),
        migrations.AlterField(
            model_name='dabpermission',
            name='content_type',
            field=models.ForeignKey(help_text='The content type this permission will apply to.', on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='content type'),
        ),
        migrations.AlterField(
            model_name='dabpermission',
            name='name',
            field=models.CharField(help_text='The name of this permission.', max_length=255, verbose_name='name'),
        ),
        migrations.AlterField(
            model_name='objectrole',
            name='content_type',
            field=models.ForeignKey(help_text='The content type of the subject of permission assignments. Duplicated from related RoleDefinition.', on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='objectrole',
            name='object_id',
            field=models.TextField(help_text='The database primary key of the subject of permission assignments.'),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='description',
            field=models.TextField(blank=True, help_text='A description of this role.'),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='managed',
            field=models.BooleanField(default=False, editable=False, help_text='Is this role managed by the system (not changeable by the users).'),
        ),
        migrations.AlterField(
            model_name='roledefinition',
            name='name',
            field=models.TextField(db_index=True, help_text='The name of this role.', unique=True),
        ),
        migrations.AlterField(
            model_name='roleevaluation',
            name='content_type_id',
            field=models.PositiveIntegerField(help_text='The related content type id.'),
        ),
        migrations.AlterField(
            model_name='roleevaluation',
            name='object_id',
            field=models.PositiveIntegerField(help_text='The id of the object that the related role gives the related permission to.'),
        ),
        migrations.AlterField(
            model_name='roleevaluationuuid',
            name='content_type_id',
            field=models.PositiveIntegerField(help_text='The related content type id.'),
        ),
        migrations.AlterField(
            model_name='roleevaluationuuid',
            name='object_id',
            field=models.UUIDField(help_text='The object UUID this role evaluation will be applied to.'),
        ),
        migrations.AlterField(
            model_name='roleteamassignment',
            name='content_type',
            field=models.ForeignKey(help_text='The content type this applies to.', null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='roleteamassignment',
            name='object_role',
            field=models.ForeignKey(editable=False, help_text='A roll-up of the fields (role_definition, content_type).', null=True, on_delete=django.db.models.deletion.CASCADE, to='dab_rbac.objectrole'),
        ),
        migrations.AlterField(
            model_name='roleteamassignment',
            name='team',
            field=models.ForeignKey(help_text='The team that receives permissions.', on_delete=django.db.models.deletion.CASCADE, related_name='role_assignments', to=settings.ANSIBLE_BASE_TEAM_MODEL),
        ),
        migrations.AlterField(
            model_name='roleuserassignment',
            name='content_type',
            field=models.ForeignKey(help_text='The content type this applies to.', null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype'),
        ),
        migrations.AlterField(
            model_name='roleuserassignment',
            name='object_role',
            field=models.ForeignKey(editable=False, help_text='A roll-up of the fields (role_definition, content_type).', null=True, on_delete=django.db.models.deletion.CASCADE, to='dab_rbac.objectrole'),
        ),
        migrations.AlterField(
            model_name='roleuserassignment',
            name='user',
            field=models.ForeignKey(help_text='The user this role is assigned to.', on_delete=django.db.models.deletion.CASCADE, related_name='role_assignments', to=settings.AUTH_USER_MODEL),
        ),
    ]



import logging

import jwt
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.utils.translation import gettext_lazy as _
from jwt.algorithms import get_default_algorithms
from jwt.exceptions import PyJWTError
from social_core.backends.open_id_connect import OpenIdConnectAuth

from ansible_base.authentication.authenticator_plugins.base import AbstractAuthenticatorPlugin, BaseAuthenticatorConfiguration
from ansible_base.authentication.social_auth import SocialAuthMixin
from ansible_base.lib.serializers.fields import BooleanField, CharField, ChoiceField, DictField, IntegerField, ListField, URLField
from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.authentication.authenticator_plugins.oidc')


DEFAULT_ALGORITHMS = get_default_algorithms()


class JWTAlgorithmListFieldValidator:

    allowed_values = list(DEFAULT_ALGORITHMS.keys())

    def __call__(self, value):
        if not all(item in self.allowed_values for item in value):
            raise ValidationError(
                _('%(value)s contains items not in the allowed list: %(allowed_values)s'),
                params={'value': value, 'allowed_values': self.allowed_values},
            )


class OpenIdConnectConfiguration(BaseAuthenticatorConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/oidc.html"

    #################################
    # Minimal params
    #################################

    OIDC_ENDPOINT = URLField(
        help_text=_("The URL for your OIDC provider including the path up to /.well-known/openid-configuration."),
        allow_null=False,
        ui_field_label=_('OIDC Provider URL'),
    )

    VERIFY_SSL = BooleanField(
        help_text=_("Verify the OIDC provider ssl certificate."),
        default=True,
        allow_null=False,
        ui_field_label=_('Verify OIDC Provider Certificate'),
    )

    KEY = CharField(
        help_text=_("The OIDC key (Client ID) from your IDP. Will also be used as the 'audience' for JWT decoding."),
        allow_null=False,
        ui_field_label=_('OIDC Key'),
    )

    SECRET = CharField(
        help_text=_("'The OIDC secret (Client Secret) from your IDP."),
        allow_null=True,
        ui_field_label=_('OIDC Secret'),
    )

    #################################
    # Additional params
    #################################

    ACCESS_TOKEN_URL = URLField(
        help_text=_("The URL to obtain an access token from the OIDC provider."),
        required=False,
        allow_null=True,
        ui_field_label=_("Access Token URL"),
    )

    ACCESS_TOKEN_METHOD = ChoiceField(
        help_text=_("The HTTP method to be used when requesting an access token. Typically 'POST' or 'GET'."),
        default="POST",
        allow_null=True,
        required=False,
        choices=['GET', 'PUT', 'POST', 'PATCH', 'DELETE'],
        ui_field_label=_("Access Token Method"),
    )

    AUTHORIZATION_URL = URLField(
        help_text=_("The URL to redirect the user for OIDC provider authorization."),
        required=False,
        allow_null=True,
        ui_field_label=_("Authorization URL"),
    )

    ID_KEY = CharField(
        help_text=_("The JSON key used to extract the user's ID from the ID token."),
        default="sub",
        allow_null=True,
        required=False,
        ui_field_label=_("ID Key"),
    )

    ID_TOKEN_ISSUER = CharField(
        help_text=_("Expected issuer ('iss') of the ID token. If set, it will be used to validate the issuer of the ID token."),
        required=False,
        allow_null=True,
        ui_field_label=_("ID Token Issuer"),
    )

    ID_TOKEN_MAX_AGE = IntegerField(
        help_text=_("The maximum allowed age (in seconds) of the ID token. Tokens older than this will be rejected."),
        default=600,
        allow_null=True,
        validators=[MinValueValidator(0)],
        ui_field_label=_('OIDC Token Max Age'),
    )

    JWT_ALGORITHMS = ListField(
        help_text=_("The algorithm(s) for decoding JWT responses from the IDP."),
        default=None,
        allow_null=True,
        validators=[JWTAlgorithmListFieldValidator()],
        ui_field_label=_('OIDC JWT Algorithm(s)'),
    )

    JWT_DECODE_OPTIONS = DictField(
        help_text=_("OIDC JWT decoding options for token validation and processing."),
        default=None,
        allow_null=True,
        ui_field_label=_('OIDC JWT Decode Options.'),
    )

    JWKS_URI = URLField(
        help_text=_("The URL to retrieve the provider's public keys for verifying JWT signatures."),
        required=False,
        allow_null=True,
        ui_field_label=_("JWKS URI"),
    )

    PUBLIC_KEY = CharField(
        help_text=_("The public key from your IDP. Only necessary if using keycloak for OIDC."),
        required=False,
        allow_null=True,
        ui_field_label=_('OIDC Public Key'),
    )

    REDIRECT_STATE = BooleanField(
        help_text=_("Enable or disable state parameter in the redirect URI. Recommended to be True for preventing CSRF attacks."),
        default=False,
        allow_null=True,
        ui_field_label=_("Redirect State"),
    )

    REVOKE_TOKEN_METHOD = ChoiceField(
        help_text=_("The HTTP method to be used when revoking an access token. Typically 'POST' or 'GET'."),
        default="GET",
        allow_null=True,
        choices=['GET', 'PUT', 'POST', 'PATCH', 'DELETE'],
        ui_field_label=_("Revoke Token Method"),
    )

    REVOKE_TOKEN_URL = URLField(
        help_text=_("The URL to revoke tokens. Used in the token revocation flow."),
        required=False,
        allow_null=True,
        ui_field_label=_("Revoke Token URL"),
    )

    RESPONSE_TYPE = CharField(
        help_text=_("The response type the OIDC endpoint should return. Common values are 'code', 'id_token' and 'token'."),
        default="code",
        allow_null=True,
        ui_field_label=_("Response Type"),
    )

    SCOPE = ListField(
        help_text=_('The authorization scope for users. Defaults to ["openid", "profile", "email"].'),
        required=False,
        allow_null=False,
        ui_field_label=_('List of OAuth2 Scope(s)'),
        default=["openid", "profile", "email"],
    )

    TOKEN_ENDPOINT_AUTH_METHOD = CharField(
        help_text=_("The authentication method to use at the token endpoint. Common values are 'client_secret_post', 'client_secret_basic'."),
        required=False,
        allow_null=True,
        ui_field_label=_("Token Endpoint Auth Method"),
    )

    USERINFO_URL = URLField(
        help_text=_("The URL to retrieve user information from the OIDC provider."),
        required=False,
        allow_null=True,
        ui_field_label=_("Userinfo URL"),
    )

    USERNAME_KEY = CharField(
        help_text=_("The JSON key used to extract the user's username from the ID token or userinfo endpoint."),
        default="preferred_username",
        required=False,
        allow_null=True,
        ui_field_label=_("Username Key"),
    )

    GROUPS_CLAIM = CharField(
        help_text=_("The JSON key used to extract the user's groups from the ID token or userinfo endpoint."),
        required=False,
        allow_null=True,
        default="Group",
        ui_field_label=_("Groups Claim"),
    )


class AuthenticatorPlugin(SocialAuthMixin, OpenIdConnectAuth, AbstractAuthenticatorPlugin):
    configuration_class = OpenIdConnectConfiguration
    type = "open_id_connect"
    logger = logger
    category = "sso"
    configuration_encrypted_fields = ['SECRET']

    @property
    def groups_claim(self):
        return self.setting('GROUPS_CLAIM')

    def extra_data(self, user, backend, response, *args, **kwargs):
        for perm in ["is_superuser", get_setting('ANSIBLE_BASE_SOCIAL_AUDITOR_FLAG')]:
            if perm in response:
                kwargs["social"].extra_data[perm] = response[perm]
        data = super().extra_data(user, backend, response, *args, **kwargs)
        return data

    def get_user_groups(self, extra_groups=[]):
        return extra_groups

    def oidc_config(self):
        # This is a copy of super without caching, which avoids data
        # from one OIDC based authenticator showing up in another
        return self.get_json(self.oidc_endpoint() + "/.well-known/openid-configuration")

    def public_key(self):
        key = self.setting("PUBLIC_KEY")
        if key:
            head = "-----BEGIN PUBLIC KEY-----"
            foot = "-----END PUBLIC KEY-----"
            return (
                "\n".join(
                    [
                        head,
                        key,
                        foot,
                    ]
                )
                if head not in key
                else key
            )
        return None

    def user_data(self, access_token, *args, **kwargs):
        """
        This function overrides the one in social auth class OpenIdConnectAuth, since
        it assumes the user info endpoint response returns plain json. Depending on
        server config it may instead be a JWT. This function notices the JWT via the
        content type header and if found, attempts to decode it using the configured
        public key and algorithm(s). None is returned to signify a failed login in the
        case of improper config or other decoding failure.
        """
        user_data = self.request(self.userinfo_url(), headers={"Authorization": f"Bearer {access_token}"})
        if user_data.headers["Content-Type"] == "application/jwt":
            # If the content type is application/jwt than we can assume that the token is encrypted. Otherwise it should be application/json
            pubkey = self.public_key()
            if not pubkey:
                logger.error(_("OIDC client sent encrypted user info response, but no public key found."))
                return None
            try:
                data = jwt.decode(
                    access_token,
                    key=pubkey,
                    algorithms=self.setting("JWT_ALGORITHMS"),
                    audience=self.setting("KEY"),
                )
                return data
            except PyJWTError as e:
                logger.error(_(f"Unable to decode user info response JWT: {e}"))
                return None
        return user_data.json()

    def get_alternative_uid(self, **kwargs):
        preferred_username = kwargs.get("response", {}).get("preferred_username", None)
        uid = kwargs.get("uid", None)

        if preferred_username != uid:
            return preferred_username

        return None



import logging

from django.apps import AppConfig
from django.conf import settings
from django.db.models import TextField, signals
from django.db.models.functions import Cast
from django.db.utils import IntegrityError

import ansible_base.lib.checks  # noqa: F401 - register checks
from ansible_base.lib.utils.db import ensure_transaction, migrations_are_complete
from ansible_base.resource_registry.utils.settings import resource_server_defined

logger = logging.getLogger('ansible_base.resource_registry.apps')


def initialize_resources(sender, **kwargs):
    from ansible_base.resource_registry.registry import get_registry

    # There isn't any evidence of this in the documentation, but it appears as though
    # Django doesn't always send the "apps" arg when it dispatches the post migrate signal
    # (https://github.com/django/django/blob/stable/4.2.x/django/core/management/sql.py#L52)
    # This seems to be the case when it is called via `django-admin flush` as well as in
    # tests that use the @pytest.mark.django_db(transaction=True) decorator.
    #
    # Since the documentation doesn't provide any clues for what do to here, we've opted
    # to rescue from scenarios where "apps" is missing by just importing the "apps" module
    # directly (which is not advised to do by the django documentation for post migrate signals
    # https://docs.djangoproject.com/en/5.0/ref/signals/#post-migrate).
    #
    # While handling this for tests doesn't matter, ignoring this function when
    # `django-admin flush` is called seems like a bad idea, since that will prevent the
    # resource types from being initialized in the database, so a direct import appears to be
    # better than doing nothing.

    if not migrations_are_complete():
        logger.info('Not running resource_registry post_migrate because migrations are incomplete')
        return

    apps = kwargs.get("apps")
    if apps is None:
        from django.apps import apps

    from ansible_base.resource_registry.models import init_resource_from_object

    Resource = apps.get_model("dab_resource_registry", "Resource")
    ResourceType = apps.get_model("dab_resource_registry", "ResourceType")
    ContentType = apps.get_model("contenttypes", "ContentType")

    logger.info("updating resource types")
    registry = get_registry()
    if registry:
        # Create resource types
        for key, resource_config in registry.get_resources().items():
            content = ContentType.objects.get_for_model(resource_config.model)

            if serializer := resource_config.managed_serializer:
                resource_type = f"shared.{serializer.RESOURCE_TYPE}"
            else:
                resource_type = f"{registry.api_config.service_type}.{content.model}"
            defaults = {"externally_managed": resource_config.externally_managed, "name": resource_type}

            try:
                ResourceType.objects.update_or_create(content_type=content, defaults=defaults)
            except IntegrityError as e:
                # if previous DAB migrations used the wrong content type id, we need to correct that now
                # to eliminate integrity errors at the end of the migration process when this function
                # gets called.
                if not ResourceType.objects.filter(name=resource_type).exists():
                    raise e
                rt = ResourceType.objects.get(name=resource_type)
                logger.warn(f"changing content_type for '{resource_type}' from '{rt.content_type.model}' to '{content.model}'")
                rt.content_type = content
                for k, v in defaults.items():
                    setattr(rt, k, v)
                rt.save()

        # Create resources
        for r_type in ResourceType.objects.all():
            resource_model = apps.get_model(r_type.content_type.app_label, r_type.content_type.model)
            resource_config = registry.get_config_for_model(resource_model)

            logger.info(f"adding unmigrated resources for {r_type.name}")

            missing_resources_qs = resource_model.objects.annotate(pk_text=Cast('pk', TextField())).exclude(
                pk_text__in=Resource.objects.filter(content_type=r_type.content_type).values("object_id")
            )

            data = []
            for obj in missing_resources_qs:
                data.append(init_resource_from_object(obj, resource_model=Resource, resource_type=r_type, resource_config=resource_config))

            Resource.objects.bulk_create(data, ignore_conflicts=True)
            r_type.save()


def proxies_of_model(cls):
    """Return models that are a proxy of cls"""
    for sub_cls in cls.__subclasses__():
        if sub_cls._meta.concrete_model is cls:
            yield sub_cls


def _should_reverse_sync():
    enabled = getattr(settings, 'RESOURCE_SERVER_SYNC_ENABLED', False)
    if enabled and (not resource_server_defined()):
        logger.debug("RESOURCE_SERVER is not configured. Reverse sync will not be enabled.")
        enabled = False
    if enabled and resource_server_defined() and ('SECRET_KEY' not in settings.RESOURCE_SERVER or not settings.RESOURCE_SERVER['SECRET_KEY']):
        logger.error("RESOURCE_SERVER['SECRET_KEY'] is not configured. Reverse sync will not be enabled.")
        enabled = False
    return enabled


def connect_resource_signals(sender, **kwargs):
    from ansible_base.resource_registry.signals import handlers

    for model in handlers.get_resource_models():
        for cls in [model, *proxies_of_model(model)]:
            # On registration, resource registry registers the concrete model
            # so we connect signals for proxies of that model, and not the other way around
            signals.post_save.connect(handlers.update_resource, sender=cls)
            signals.post_delete.connect(handlers.remove_resource, sender=cls)

            if _should_reverse_sync():
                signals.pre_save.connect(handlers.decide_to_sync_update, sender=cls)
                signals.post_save.connect(handlers.sync_to_resource_server_post_save, sender=cls)
                signals.pre_delete.connect(handlers.sync_to_resource_server_pre_delete, sender=cls)

                # Wrap save() in a transaction and sync to resource server
                cls._original_save = cls.save

                # Avoid late binding issues
                def save(self, *args, _original_save=cls._original_save, **kwargs):
                    with ensure_transaction():
                        _original_save(self, *args, **kwargs)

                cls.save = save

                # Wrap delete() in a transaction and remove from resource server
                cls._original_delete = cls.delete

                # Avoid late binding issues
                def delete(self, *args, _original_delete=cls._original_delete, **kwargs):
                    with ensure_transaction():
                        _original_delete(self, *args, **kwargs)

                cls.delete = delete


def disconnect_resource_signals(sender, **kwargs):
    from ansible_base.resource_registry.signals import handlers

    for model in handlers.get_resource_models():
        for cls in [model, *proxies_of_model(model)]:
            signals.post_save.disconnect(handlers.update_resource, sender=cls)
            signals.post_delete.disconnect(handlers.remove_resource, sender=cls)

            signals.pre_save.disconnect(handlers.decide_to_sync_update, sender=cls)
            signals.post_save.disconnect(handlers.sync_to_resource_server_post_save, sender=cls)
            signals.pre_delete.disconnect(handlers.sync_to_resource_server_pre_delete, sender=cls)

            if hasattr(cls, '_original_save'):
                cls.save = cls._original_save
                del cls._original_save

            if hasattr(cls, '_original_delete'):
                cls.delete = cls._original_delete
                del cls._original_delete


class ResourceRegistryConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.resource_registry'
    label = 'dab_resource_registry'
    verbose_name = 'Service resources API'

    def ready(self):
        connect_resource_signals(sender=None)
        signals.pre_migrate.connect(disconnect_resource_signals, sender=self)
        signals.post_migrate.connect(initialize_resources, sender=self)
        signals.post_migrate.connect(connect_resource_signals, sender=self)



from django.contrib import admin

from ansible_base.lib.admin import ReadOnlyAdmin
from ansible_base.rbac.models import ObjectRole, RoleDefinition, RoleEvaluation, RoleTeamAssignment, RoleUserAssignment

admin.site.register(RoleDefinition)
# TODO: assignments will still not be functional in the admin pages without custom logic
admin.site.register(RoleUserAssignment)
admin.site.register(RoleTeamAssignment)
admin.site.register(ObjectRole, ReadOnlyAdmin)
admin.site.register(RoleEvaluation, ReadOnlyAdmin)



import logging
import os

from crum import get_current_user
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import ValidationError

from ansible_base.resource_registry.models import Resource, service_id
from ansible_base.resource_registry.rest_client import ResourceRequestBody, get_resource_server_client

logger = logging.getLogger('ansible_base.resource_registry.utils.sync_to_resource_server')


def sync_to_resource_server(instance, action, ansible_id=None):
    """
    Use the resource server API to sync the resource across.

    When action is "delete", the ansible_id is required, because by the time we
    get here, we've already deleted the object and its resource. So we can't
    pull the ansible_id from the resource object. It's on the caller to pull
    the ansible_id from the object before deleting it.

    For all other actions, ansible_id is ignored and retrieved from the resource
    object. (For create, the resource is expected to exist before calling this
    function.)
    """

    sync_disabled = os.environ.get('ANSIBLE_REVERSE_RESOURCE_SYNC', 'true').lower() == 'false'
    if sync_disabled:
        logger.info(f"Skipping sync of resource {instance} because $ANSIBLE_REVERSE_RESOURCE_SYNC is 'false'")
        return

    # This gets set in in signals.handlers.decide_to_sync_update() sometimes.
    skip_sync = getattr(instance, '_skip_reverse_resource_sync', False)
    if skip_sync:
        # Avoid an infinite loop by not syncing resources that came from the resource server.
        # Or avoid syncing unnecessarily, when a synced field hasn't changed.
        logger.info(f"Skipping sync of resource {instance}")
        return

    if action != "delete" and ansible_id is not None:
        raise Exception("ansible_id should not be provided for create/update actions")
    elif action == "delete" and ansible_id is None:
        raise Exception("ansible_id should be provided for delete actions")

    try:
        resource = Resource.get_resource_for_object(instance)
    except Resource.DoesNotExist:
        logger.error(f"Resource {instance} does not have a resource")
        return

    if str(resource.service_id) == service_id() and action == "update":
        # Don't sync if we're updating a resource that isn't owned by the resource server yet.
        logger.info(f"Skipping sync of resource {instance} because its service_id is local")
        return

    user_ansible_id = None
    user = get_current_user()
    if user:
        # If we have a user, try to get their ansible_id and sync as them.
        # If they don't have one some how, or if we don't have a user, sync with None and
        # let the resource server decide what to do.
        try:
            user_resource = Resource.get_resource_for_object(user)
            user_ansible_id = user_resource.ansible_id
        except (Resource.DoesNotExist, AttributeError):
            logger.error(f"User {user} does not have a resource")
            pass
    else:
        logger.error("No user found, syncing to resource server with jwt_user_id=None")

    client = get_resource_server_client(
        settings.RESOURCE_SERVICE_PATH,
        jwt_user_id=user_ansible_id,
        raise_if_bad_request=True,
    )

    if action != "delete":
        ansible_id = resource.ansible_id

    resource_type = resource.content_type.resource_type
    data = resource_type.serializer_class(instance).data
    body = ResourceRequestBody(
        resource_type=resource_type.name,
        ansible_id=ansible_id,
        resource_data=data,
    )

    try:
        if action == "create":
            response = client.create_resource(body)
            json = response.json()
            if isinstance(json, dict):  # This 'isinstance' check is mainly for tests... to avoid getting here with mock
                # The service_id is saved here because we don't have a local reference to it anywhere.
                # The ansible_id is saved here because of the following scenario:
                #    Service A and Service B both try to create resource C at the same time.
                #    Service A's request arrives first and receives the lock on the DB to create the resource and creates resource C with ID=1
                #    Once the lock on the DB is released, Service B's request comes through.
                #        Because this endpoint does a create or update operation, Service B modifies resource C, and sets its ID=2.
                #    Now resource C is out of sync. On the resource server and service B, the ID=2, but on service A the ID is now 1.
                #    Fixing this problem is fairly easy. We just let the resource server set the ansible ID of the resource,
                #        rather than let each service pick their own random UUID.
                resource.service_id = json['service_id']
                resource.ansible_id = json['ansible_id']
                resource.save()
        elif action == "update":
            client.update_resource(ansible_id, body)
        elif action == "delete":
            client.delete_resource(ansible_id)
    except Exception as e:
        logger.exception(f"Failed to sync {action} of resource {instance} ({ansible_id}) to resource server: {e}")
        raise ValidationError(_("Failed to sync resource to resource server")) from e



from typing import Optional

from django.contrib.auth import get_user_model

from ansible_base.authentication.models import Authenticator

User = get_user_model()


def is_external_account(user: User) -> Optional[Authenticator]:
    """
    Determines whether the user is associated with any external
    login source. If they are, return the source. Otherwise, None.

    :param user: The user to test
    :return: If the user is associated with any external login source, return it (the first, if multiple)
             Otherwise, return None
    """
    authenticator_users = user.authenticator_users.all()
    local = 'ansible_base.authentication.authenticator_plugins.local'
    for auth_user in authenticator_users:
        if auth_user.provider.type != local:
            return auth_user.provider

    return None



from copy import copy
from typing import Optional

from ansible_base.lib.cache.fallback_cache import FALLBACK_CACHE, PRIMARY_CACHE

#
# If you are adding a new dynamic setting:
#     Please be sure to modify pyproject.toml with your new settings in tool.setuptools.dynamic
#     Add a new requirements/requirements_<section>.in /even if its an empty file/
#


DEFAULT_AUTH_GROUP = 'auth.Group'
DEFAULT_SPECTACULAR_SETTINGS = {
    'TITLE': 'Open API',
    'DESCRIPTION': 'Open API',
    'VERSION': 'v1',
    'SCHEMA_PATH_PREFIX': '/api/v1/',
    'COMPONENT_NO_READ_ONLY_REQUIRED': True,
}
DEFAULT_ANSIBLE_BASE_AUTH = "ansible_base.authentication.backend.AnsibleBaseAuth"
DEFAULT_ANSIBLE_BASE_JWT_CONSUMER_APP_NAME = "ansible_base.jwt_consumer"
DEFAULT_ANSIBLE_BASE_RBAC_APP_NAME = "ansible_base.rbac"
DEFAULT_OAUTH2_APPLICATION_MODEL = 'dab_oauth2_provider.OAuth2Application'
DEFAULT_OAUTH2_ACCESS_TOKEN = 'dab_oauth2_provider.OAuth2AccessToken'
DEFAULT_TEMPLATE_BACKEND = 'django.template.backends.django.DjangoTemplates'


def get_dab_settings(
    installed_apps: list[str],
    rest_framework: Optional[dict] = None,
    spectacular_settings: Optional[dict] = None,
    authentication_backends: Optional[list[str]] = None,
    middleware: Optional[list[str]] = None,
    oauth2_provider: Optional[dict] = None,
    caches: Optional[dict] = None,
    templates: Optional[list[dict]] = None,
) -> dict:  # pragma: no cover
    """
    This function is deprecated and will be removed in a future version.
    Please use `ansible_base.lib.dynamic_config.factory` instead.
    """
    settings = {
        "INSTALLED_APPS": installed_apps,
        "REST_FRAMEWORK": rest_framework or {},
        "MIDDLEWARE": middleware or [],
    }
    if spectacular_settings is not None:
        settings["SPECTACULAR_SETTINGS"] = spectacular_settings
    if authentication_backends is not None:
        settings["AUTHENTICATION_BACKENDS"] = authentication_backends
    if oauth2_provider is not None:
        settings["OAUTH2_PROVIDER"] = oauth2_provider
    if caches is not None:
        settings["CACHES"] = caches
    if templates is not None:
        settings["TEMPLATES"] = templates
    return get_mergeable_dab_settings(settings)


def get_mergeable_dab_settings(settings: dict) -> dict:  # NOSONAR
    """This function takes a settings dict and returns a dict of DAB settings
     that can be merged into the Dynaconf settings or directly set to any module or object.

    TODO: In a future implementation this function will defer each section to specific app logic.
    each app will have its own get_defaults function that will return the settings to be merged.
    e.g:
        from ansible_base.resource_registry.settings import get_defaults
        resource_registry_settings = get_defaults(settings)
        dab_data.update(resource_registry_settings)

    The call here can be explicit, or dynamic based on the INSTALLED_APPS list.
    """
    dab_data = {}

    installed_apps: list = copy(settings["INSTALLED_APPS"])
    middleware: list = copy(settings["MIDDLEWARE"])
    rest_framework: dict = copy(settings["REST_FRAMEWORK"])

    oauth2_provider: dict = copy(settings.get("OAUTH2_PROVIDER", {}))
    templates: list = copy(settings.get("TEMPLATES", []))
    authentication_backends: list = copy(settings.get("AUTHENTICATION_BACKENDS", []))
    spectacular_settings = settings.get('SPECTACULAR_SETTINGS', {})

    # The org and team abstract models cause errors if not set, even if not used
    if settings.get('ANSIBLE_BASE_TEAM_MODEL') is None:
        dab_data['ANSIBLE_BASE_TEAM_MODEL'] = DEFAULT_AUTH_GROUP
    if settings.get('ANSIBLE_BASE_ORGANIZATION_MODEL') is None:
        dab_data['ANSIBLE_BASE_ORGANIZATION_MODEL'] = DEFAULT_AUTH_GROUP

    # This is needed for the rest_filters app, but someone may use the filter class
    # without enabling the ansible_base.rest_filters app explicitly
    # we also apply this to views from other apps so we should always define it
    if settings.get('ANSIBLE_BASE_REST_FILTERS_RESERVED_NAMES') is None:
        dab_data['ANSIBLE_BASE_REST_FILTERS_RESERVED_NAMES'] = (
            'page',
            'page_size',
            'format',
            'order',
            'order_by',
            'search',
            'type',
            'host_filter',
            'count_disabled',
            'no_truncate',
            'limit',
            'validate',
        )

    # SPECTACULAR SETTINGS
    if 'ansible_base.api_documentation' in installed_apps:
        rest_framework.setdefault('DEFAULT_SCHEMA_CLASS', 'drf_spectacular.openapi.AutoSchema')

        if 'drf_spectacular' not in installed_apps:
            installed_apps.append('drf_spectacular')

        for key, value in DEFAULT_SPECTACULAR_SETTINGS.items():
            if key not in spectacular_settings:
                spectacular_settings[key] = value

    # General, factual, constant of all filters that ansible_base.rest_filters ships
    dab_data['ANSIBLE_BASE_ALL_REST_FILTERS'] = (
        'ansible_base.rest_filters.rest_framework.type_filter_backend.TypeFilterBackend',
        'ansible_base.rest_filters.rest_framework.field_lookup_backend.FieldLookupBackend',
        'rest_framework.filters.SearchFilter',
        'ansible_base.rest_filters.rest_framework.order_backend.OrderByBackend',
    )
    if 'ansible_base.rest_filters' in installed_apps:
        rest_framework['DEFAULT_FILTER_BACKENDS'] = dab_data['ANSIBLE_BASE_ALL_REST_FILTERS']
    else:
        # Explanation - these are the filters for views provided by DAB like /authenticators/
        # we want them to be enabled by default _even if_ the rest_filters app is not used
        # so that clients have consistency, but if an app wants to turn them off, they can.
        # these will be combined with the actual DRF defaults in our base view
        dab_data['ANSIBLE_BASE_CUSTOM_VIEW_FILTERS'] = dab_data['ANSIBLE_BASE_ALL_REST_FILTERS']

    if 'ansible_base.authentication' in installed_apps:
        if 'social_django' not in installed_apps:
            installed_apps.append('social_django')
        if DEFAULT_ANSIBLE_BASE_AUTH not in authentication_backends:
            authentication_backends.append(DEFAULT_ANSIBLE_BASE_AUTH)

        middleware_classes = [
            'ansible_base.authentication.middleware.SocialExceptionHandlerMiddleware',
            'ansible_base.authentication.middleware.AuthenticatorBackendMiddleware',
        ]
        for mw in middleware_classes:
            if mw not in middleware:
                try:
                    index = middleware.index('django.contrib.auth.middleware.AuthenticationMiddleware')
                    middleware.insert(index, mw)
                except ValueError:
                    middleware.append(mw)

        drf_authentication_class = 'ansible_base.authentication.session.SessionAuthentication'
        rest_framework.setdefault('DEFAULT_AUTHENTICATION_CLASSES', [])
        if drf_authentication_class not in rest_framework['DEFAULT_AUTHENTICATION_CLASSES']:
            rest_framework['DEFAULT_AUTHENTICATION_CLASSES'].insert(0, drf_authentication_class)

        if settings.get('ANSIBLE_BASE_AUTHENTICATOR_CLASS_PREFIXES') is None:
            dab_data['ANSIBLE_BASE_AUTHENTICATOR_CLASS_PREFIXES'] = ["ansible_base.authentication.authenticator_plugins"]

        dab_data['SOCIAL_AUTH_PIPELINE'] = (
            'social_core.pipeline.social_auth.social_details',
            'social_core.pipeline.social_auth.social_uid',
            'social_core.pipeline.social_auth.auth_allowed',
            'social_core.pipeline.social_auth.social_user',
            'ansible_base.authentication.utils.authentication.determine_username_from_uid_social',
            'social_core.pipeline.user.create_user',
            'social_core.pipeline.social_auth.associate_user',
            'social_core.pipeline.social_auth.load_extra_data',
            'social_core.pipeline.user.user_details',
            'ansible_base.authentication.social_auth.create_user_claims_pipeline',
        )
        dab_data['SOCIAL_AUTH_STORAGE'] = "ansible_base.authentication.social_auth.AuthenticatorStorage"
        dab_data['SOCIAL_AUTH_STRATEGY'] = "ansible_base.authentication.social_auth.AuthenticatorStrategy"
        dab_data['SOCIAL_AUTH_LOGIN_REDIRECT_URL'] = "/"
        dab_data['ANSIBLE_BASE_SOCIAL_AUDITOR_FLAG'] = "is_system_auditor"
        # URL to send users when social auth login fails
        dab_data['LOGIN_ERROR_URL'] = "/?auth_failed"

    if 'ansible_base.rest_pagination' in installed_apps:
        rest_framework.setdefault('DEFAULT_PAGINATION_CLASS', 'ansible_base.rest_pagination.DefaultPaginator')

    if DEFAULT_ANSIBLE_BASE_JWT_CONSUMER_APP_NAME in installed_apps and DEFAULT_ANSIBLE_BASE_RBAC_APP_NAME not in installed_apps:
        installed_apps.append(DEFAULT_ANSIBLE_BASE_RBAC_APP_NAME)

    if (DEFAULT_ANSIBLE_BASE_JWT_CONSUMER_APP_NAME in installed_apps) or (DEFAULT_ANSIBLE_BASE_RBAC_APP_NAME in installed_apps):
        dab_data['ANSIBLE_BASE_JWT_MANAGED_ROLES'] = ["Platform Auditor", "Organization Admin", "Organization Member", "Team Admin", "Team Member"]

    rbac_defaults = {
        # The settings-based specification of managed roles from DAB RBAC vendored ones
        'ANSIBLE_BASE_MANAGED_ROLE_REGISTRY': {},
        # Permissions a user will get when creating a new item
        'ANSIBLE_BASE_CREATOR_DEFAULTS': ['add', 'change', 'delete', 'view'],
        # Permissions API will check for related items, think PATCH/PUT
        # This is a precedence order, so first action related model has will be used
        'ANSIBLE_BASE_CHECK_RELATED_PERMISSIONS': ['use', 'change', 'view'],
        # If a role does not already exist that can give those object permissions
        # then the system must create one, this is used for naming the auto-created role
        'ANSIBLE_BASE_ROLE_CREATOR_NAME': '{obj._meta.model_name}-creator-permission',
        # Require view permission in roles containing any other permission
        # this requirement does not apply to models that do not have view permission
        'ANSIBLE_BASE_ROLES_REQUIRE_VIEW': True,
        # Require change permission to get delete permission
        'ANSIBLE_BASE_DELETE_REQUIRE_CHANGE': True,
        # Specific feature enablement bits
        # For assignments
        'ANSIBLE_BASE_ALLOW_TEAM_PARENTS': True,
        'ANSIBLE_BASE_ALLOW_TEAM_ORG_PERMS': True,
        'ANSIBLE_BASE_ALLOW_TEAM_ORG_MEMBER': False,
        'ANSIBLE_BASE_ALLOW_TEAM_ORG_ADMIN': True,
        # For role definitions
        'ANSIBLE_BASE_ALLOW_CUSTOM_ROLES': True,
        'ANSIBLE_BASE_ALLOW_CUSTOM_TEAM_ROLES': False,
        # Allows managing singleton permissions
        'ANSIBLE_BASE_ALLOW_SINGLETON_USER_ROLES': False,
        'ANSIBLE_BASE_ALLOW_SINGLETON_TEAM_ROLES': False,
        'ANSIBLE_BASE_ALLOW_SINGLETON_ROLES_API': True,
        # Pass ignore_conflicts=True for bulk_create calls for role evaluations
        # this should be fine to resolve cross-process conflicts as long as
        # directionality is the same - adding or removing permissions
        # A value of False would result in more errors but be more conservative
        'ANSIBLE_BASE_EVALUATIONS_IGNORE_CONFLICTS': True,
        # User flags that can grant permission before consulting roles
        'ANSIBLE_BASE_BYPASS_SUPERUSER_FLAGS': ['is_superuser'],
        'ANSIBLE_BASE_BYPASS_ACTION_FLAGS': {},
        # Save RoleEvaluation entries for child permissions on parent models
        # ex: organization roles giving view_inventory permission will save
        # entries mapping that permission to the assignment's organization
        'ANSIBLE_BASE_CACHE_PARENT_PERMISSIONS': False,
        # API clients can assign users and teams roles for shared resources
        'ALLOW_LOCAL_RESOURCE_MANAGEMENT': True,
        # API clients can assign roles provided by the JWT
        # this should only be left as True for testing purposes
        # TODO: change this default to False
        'ALLOW_LOCAL_ASSIGNING_JWT_ROLES': True,
        # API clients can create custom roles that change shared resources
        'ALLOW_SHARED_RESOURCE_CUSTOM_ROLES': False,
        'MANAGE_ORGANIZATION_AUTH': True,
        # Alternative to permission_registry.register
        'ANSIBLE_BASE_RBAC_MODEL_REGISTRY': {},
        'ORG_ADMINS_CAN_SEE_ALL_USERS': True,
    }
    if DEFAULT_ANSIBLE_BASE_RBAC_APP_NAME in installed_apps:
        for key, value in rbac_defaults.items():
            if settings.get(key) is None:
                dab_data[key] = value

    resource_registry_defaults = {
        # Sync local changes to the resource server
        # This will not do anything if RESOURCE_SERVER is not defined
        'RESOURCE_SERVER_SYNC_ENABLED': True,
        # The API path on the resource server to use to update resources
        'RESOURCE_SERVICE_PATH': "/api/gateway/v1/service-index/",
        # Disable legacy SSO by default
        'ENABLE_SERVICE_BACKED_SSO': False,
    }
    if 'ansible_base.resource_registry' in installed_apps:
        for key, value in resource_registry_defaults.items():
            if settings.get(key) is None:
                dab_data[key] = value

    if 'ansible_base.oauth2_provider' in installed_apps:
        if 'oauth2_provider' not in installed_apps:
            installed_apps.append('oauth2_provider')

        oauth2_provider.setdefault('ACCESS_TOKEN_EXPIRE_SECONDS', 31536000000)
        oauth2_provider.setdefault('AUTHORIZATION_CODE_EXPIRE_SECONDS', 600)
        oauth2_provider.setdefault('REFRESH_TOKEN_EXPIRE_SECONDS', 2628000)
        # For compat with awx, we don't require PKCE, but the new version
        # of DOT that we are using requires it by default.
        oauth2_provider.setdefault('PKCE_REQUIRED', False)

        oauth2_provider['OAUTH2_BACKEND_CLASS'] = 'ansible_base.oauth2_provider.authentication.OAuthLibCore'
        oauth2_provider['APPLICATION_MODEL'] = DEFAULT_OAUTH2_APPLICATION_MODEL
        oauth2_provider['ACCESS_TOKEN_MODEL'] = DEFAULT_OAUTH2_ACCESS_TOKEN

        rest_framework.setdefault('DEFAULT_AUTHENTICATION_CLASSES', [])
        oauth2_authentication_class = 'ansible_base.oauth2_provider.authentication.LoggedOAuth2Authentication'
        if oauth2_authentication_class not in rest_framework['DEFAULT_AUTHENTICATION_CLASSES']:
            rest_framework['DEFAULT_AUTHENTICATION_CLASSES'].insert(0, oauth2_authentication_class)

        # These have to be defined for the migration to function
        dab_data['OAUTH2_PROVIDER_APPLICATION_MODEL'] = DEFAULT_OAUTH2_APPLICATION_MODEL
        dab_data['OAUTH2_PROVIDER_ACCESS_TOKEN_MODEL'] = DEFAULT_OAUTH2_ACCESS_TOKEN
        dab_data['OAUTH2_PROVIDER_REFRESH_TOKEN_MODEL'] = "dab_oauth2_provider.OAuth2RefreshToken"
        dab_data['OAUTH2_PROVIDER_ID_TOKEN_MODEL'] = "dab_oauth2_provider.OAuth2IDToken"

        dab_data['ALLOW_OAUTH2_FOR_EXTERNAL_USERS'] = False
        dab_data['ANSIBLE_BASE_OAUTH2_PROVIDER_PERMISSIONS_CHECK_DEFAULT_IGNORED_VIEWS'] = []

    # Ensure proper configuration for fallback cache
    if (caches := settings.get("CACHES")) is not None:
        dab_data['CACHES'] = caches
        default_backend = caches.get('default', {}).get('BACKEND', '')
        if default_backend == 'ansible_base.lib.cache.fallback_cache.DABCacheWithFallback':
            # Ensure primary and fallback are defined
            if PRIMARY_CACHE not in caches or FALLBACK_CACHE not in caches:
                raise RuntimeError(f'Cache definitions with the keys {PRIMARY_CACHE} and {FALLBACK_CACHE} must be defined when DABCacheWithFallback is used.')

    if 'ansible_base.feature_flags' in installed_apps:
        if "flags" not in installed_apps:
            installed_apps.append('flags')

        found_template_backend = False
        template_context_processor = 'django.template.context_processors.request'
        # Look through all of the tmplates
        for template in templates:
            # If this template has the BACKEND we care about...
            if template['BACKEND'] == DEFAULT_TEMPLATE_BACKEND:
                found_template_backend = True
                # Look through all of its context processors
                found_context_processor = False
                for context_processor in template['OPTIONS']['context_processors']:
                    if context_processor == template_context_processor:
                        found_context_processor = True
                        break
                # If we didn't find the context processor we care about append it
                if not found_context_processor:
                    template['OPTIONS']['context_processors'].append(template_context_processor)

        # If we never even found the backend, add one
        if not found_template_backend:
            templates.append(
                {
                    'BACKEND': DEFAULT_TEMPLATE_BACKEND,
                    'OPTIONS': {'context_processors': [template_context_processor]},
                }
            )

    # Set back only changed base keys so the inspect history is accurate
    if installed_apps != settings["INSTALLED_APPS"]:
        dab_data["INSTALLED_APPS"] = installed_apps
    if middleware != settings["MIDDLEWARE"]:
        dab_data["MIDDLEWARE"] = middleware
    if rest_framework != settings["REST_FRAMEWORK"]:
        dab_data["REST_FRAMEWORK"] = rest_framework
    if oauth2_provider != settings.get("OAUTH2_PROVIDER", {}):
        dab_data["OAUTH2_PROVIDER"] = oauth2_provider
    if templates != settings.get("TEMPLATES", []):
        dab_data["TEMPLATES"] = templates
    if authentication_backends != settings.get("AUTHENTICATION_BACKENDS", []):
        dab_data["AUTHENTICATION_BACKENDS"] = authentication_backends
    if spectacular_settings != settings.get('SPECTACULAR_SETTINGS', {}):
        dab_data['SPECTACULAR_SETTINGS'] = spectacular_settings

    return dab_data



from django.apps import AppConfig


class RestPaginationConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.rest_pagination'
    label = 'dab_rest_pagination'



from django.utils.translation import gettext_lazy as _

from ansible_base.authentication.authenticator_plugins.base import BaseAuthenticatorConfiguration
from ansible_base.lib.serializers.fields import CharField, ListField, URLField


class GithubConfiguration(BaseAuthenticatorConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/github.html"

    CALLBACK_URL = URLField(
        help_text=_(
            'Provide this URL as the callback URL for your application as part of your registration process. Refer to the documentation for more detail.'
        ),
        allow_null=False,
        required=False,
        ui_field_label=_('Github Oauth2 Callback URL'),
    )

    KEY = CharField(
        help_text=_('The OAuth2 key (Client ID) from your GitHub developer application.'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 Key'),
    )

    SECRET = CharField(
        help_text=_('The OAuth2 secret (Client Secret) from your GitHub developer application.'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 Secret'),
    )


class GithubOrganizationConfiguration(GithubConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/github.html#github-for-organizations"

    SCOPE = ListField(
        help_text=_('The authorization scope for users. Defaults to "read:org".'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 Scope'),
        default=['read:org'],
    )

    NAME = CharField(
        help_text=_('The OAuth2 organization name.'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 Organization Name'),
    )


class GithubTeamConfiguration(GithubConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/github.html#github-for-teams"

    SCOPE = ListField(
        help_text=_('The authorization scope for users. Defaults to "read:org".'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 Scope'),
        default=['read:org'],
    )

    ID = CharField(
        help_text=_('The OAuth2 team ID.'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 Team ID'),
    )


###############################################################################
#   ENTERPRISE
###############################################################################


class GithubEnterpriseConfiguration(GithubConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/github_enterprise.html"

    URL = URLField(
        help_text=_('The base url for the GithHub enterprise instance.'),
        allow_null=False,
        ui_field_label=_('Base URL'),
    )

    API_URL = URLField(
        help_text=_('The base url for the GithHub enterprise instance.'),
        allow_null=False,
        ui_field_label=_('Github OAuth2 Enterprise API URL'),
    )


class GithubEnterpriseOrgConfiguration(GithubEnterpriseConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/github_enterprise.html#github-enterprise-for-organizations"

    NAME = CharField(
        help_text=_('The OAuth2 key (Client ID) from your GitHub developer application.'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 Enterprise Org Name'),
    )


class GithubEnterpriseTeamConfiguration(GithubEnterpriseConfiguration):
    documentation_url = "https://python-social-auth.readthedocs.io/en/latest/backends/github_enterprise.html#github-enterprise-for-teams"

    ID = CharField(
        help_text=_('The OAuth2 key (Client ID) from your GitHub developer application.'),
        allow_null=False,
        ui_field_label=_('GitHub OAuth2 team ID'),
    )



from django.urls import include, path
from rest_framework import routers

from ansible_base.activitystream import views

router = routers.SimpleRouter()
router.register(
    'activitystream',
    views.EntryReadOnlyViewSet,
    basename='activitystream',
)

api_version_urls = [path('', include(router.urls))]



# Generated by Django 4.2.17 on 2025-02-04 20:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('dab_authentication', '0016_alter_authenticatoruser_access_allowed_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='authenticator',
            name='slug',
            field=models.SlugField(default=None, editable=False, help_text='An immutable identifier for the authenticator; used to generate the sso uri for sso authenticator types', max_length=1024, unique=True),
        ),
    ]



from pathlib import Path

from requests import Response

from ansible_base.resource_registry.rest_client import ResourceAPIClient


def copy_fixture(copies=1):
    """
    Decorator to create 'copies' copies of a fixture.

    The copies will be named func_1, func_2, ..., func_n in the same module as
    the original fixture.
    """

    def wrapper(func):
        if '_pytestfixturefunction' not in dir(func):
            raise TypeError(f"Can't apply copy_fixture to {func.__name__} because it is not a fixture. HINT: @copy_fixture must be *above* @pytest.fixture")

        module_name = func.__module__
        module = __import__(module_name, fromlist=[''])

        for i in range(copies):
            new_name = f"{func.__name__}_{i + 1}"
            setattr(module, new_name, func)
        return func

    return wrapper


def delete_authenticator(authenticator):
    from ansible_base.authentication.models import AuthenticatorUser
    from ansible_base.lib.utils.models import get_system_username

    for au in AuthenticatorUser.objects.filter(provider=authenticator):
        try:
            # The tests are very sensitive to the system user being removed so we won't delete that user
            if au.username != get_system_username()[0]:
                au.user.delete()
        except Exception:
            # Its possible that something else already deleted the user if a user was multi linked somehow
            pass
        au.delete()
    authenticator.delete()


class StaticResourceAPIClient(ResourceAPIClient):
    """A testing API client that reads response router attribute or static files."""

    router = {}
    # Route is used to force a certain status,response for a route
    # It has to be a mutable default but the fixture instantiates one for
    # each test.

    def _make_request(self, method, path, data=None, params=None, stream=False):
        response = Response()
        response.status_code = 200
        response.encoding = "utf-8"

        if path in self.router:
            response.status_code = self.router[path]["status_code"]
            response._content = self.router[path]["content"]
            return response

        content_file_path = Path(self.base_url) / path / "response"

        try:
            response._content = content_file_path.read_bytes()
        except FileNotFoundError:
            response.status_code = 404

        return response



import inspect
import logging
from os.path import dirname, join
from urllib.parse import urlparse, urlunsplit

from django.http import HttpResponse, HttpResponseNotFound
from django.template import Context, Template
from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions, permissions

from ansible_base.lib.utils.settings import get_setting
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView

logger = logging.getLogger('ansible_base.jwt_consumer.views')


class PlatformUIRedirectView(AnsibleBaseDjangoAppApiView):
    authentication_classes = []
    permission_classes = (permissions.AllowAny,)
    metadata_class = None
    exception_class = exceptions.APIException

    name = _('Platform UI Redirect')

    def finalize_response(self, request, response, *args, **kwargs):
        url = get_setting("ANSIBLE_BASE_JWT_REDIRECT_URI", get_setting("ANSIBLE_BASE_JWT_KEY", None))
        service_type = get_setting("ANSIBLE_BASE_JWT_REDIRECT_TYPE", "unknown")
        if not url:
            return HttpResponseNotFound()

        url_parts = urlparse(url)
        base_url = urlunsplit((url_parts[0], url_parts[1], '', '', ''))
        context = Context({'redirect_url': base_url, 'service': service_type})
        try:
            filename = join(dirname(inspect.getfile(self.__class__)), 'redirect.html')
            with open(filename, 'r') as f:
                redirect_template = f.read()
        except Exception as e:
            logger.error(f"Failed to load redirect.html: {e}")
            return HttpResponseNotFound()

        template = Template(redirect_template)
        return HttpResponse(template.render(context))



from django.core.exceptions import FieldDoesNotExist, FieldError
from rest_framework.exceptions import ParseError
from rest_framework.filters import BaseFilterBackend

from ansible_base.rest_filters.utils import get_all_field_names, get_field_from_path


class OrderByBackend(BaseFilterBackend):
    """
    Filter to apply ordering based on query string parameters.
    """

    def filter_queryset(self, request, queryset, view):
        try:
            order_by = None
            for key, value in request.query_params.items():
                if key in ('order', 'order_by'):
                    order_by = value
                    if ',' in value:
                        order_by = value.split(',')
                    else:
                        order_by = (value,)
            default_order_by = self.get_default_ordering(view)
            # glue the order by and default order by together so that the default is the backup option
            order_by = list(order_by or []) + list(default_order_by or [])
            if order_by:
                order_by = self._validate_ordering_fields(queryset.model, order_by)
                # Special handling of the type field for ordering. In this
                # case, we're not sorting exactly on the type field, but
                # given the limited number of views with multiple types,
                # sorting on polymorphic_ctype.model is effectively the same.
                new_order_by = []
                if 'polymorphic_ctype' in get_all_field_names(queryset.model):
                    for field in order_by:
                        if field == 'type':
                            new_order_by.append('polymorphic_ctype__model')
                        elif field == '-type':
                            new_order_by.append('-polymorphic_ctype__model')
                        else:
                            new_order_by.append(field)
                else:
                    for field in order_by:
                        if field not in ('type', '-type'):
                            new_order_by.append(field)
                queryset = queryset.order_by(*new_order_by)
            return queryset
        except FieldError as e:
            # Return a 400 for invalid field names.
            raise ParseError(*e.args)

    def get_default_ordering(self, view):
        ordering = getattr(view, 'ordering', None)
        if isinstance(ordering, str):
            return (ordering,)
        return ordering

    def _validate_ordering_fields(self, model, order_by):
        for field_name in order_by:
            # strip off the negation prefix `-` if it exists
            prefix = ''
            path = field_name
            if field_name[0] == '-':
                prefix = field_name[0]
                path = field_name[1:]
            try:
                field, new_path = get_field_from_path(model, path)
                new_path = '{}{}'.format(prefix, new_path)
            except (FieldError, FieldDoesNotExist) as e:
                raise ParseError(e.args[0])
            yield new_path



# Generated by Django 4.2.16 on 2024-11-22 21:25

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('dab_oauth2_provider', '0006_alter_oauth2accesstoken_created_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='application',
            field=models.ForeignKey(blank=True, help_text='The related application. If None, this is a user token instead of an application token.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='access_tokens', to=settings.OAUTH2_PROVIDER_APPLICATION_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='description',
            field=models.TextField(blank=True, default='', help_text='A description for this token.'),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='last_used',
            field=models.DateTimeField(default=None, editable=False, help_text='A timestamp of when this token was last used.', null=True),
        ),
        migrations.AlterField(
            model_name='oauth2accesstoken',
            name='token',
            field=models.CharField(help_text='The generated token value.', max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='description',
            field=models.TextField(blank=True, default='', help_text='A description of this application.'),
        ),
        migrations.AlterField(
            model_name='oauth2application',
            name='user',
            field=models.ForeignKey(blank=True, help_text='The user who created the application.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='applications', to=settings.AUTH_USER_MODEL),
        ),
        migrations.AlterField(
            model_name='oauth2refreshtoken',
            name='token',
            field=models.CharField(help_text='The refresh token value.', max_length=255),
        ),
    ]



from typing import Optional

from django.contrib.auth.models import AbstractUser

from ansible_base.authentication.authenticator_plugins.utils import get_authenticator_plugin
from ansible_base.authentication.models import AuthenticatorUser
from ansible_base.lib.utils.models import is_system_user


def can_user_change_password(user: Optional[AbstractUser]) -> bool:
    """
    See if the given user is allowed to change their password.
    True if they are authenticated from the `local` authenticator
    False otherwise.
    The system user can never change their password
    """
    if user is None or is_system_user(user):
        # If we didn't actually get a user we can't say they can change their password
        # Or if we are the system user, we can not change our password ever
        return False

    auth_users = AuthenticatorUser.objects.filter(user=user)
    if auth_users.count() == 0:
        # If the user has no associations we can set a password for them so they can login through the local authenticator
        return True

    for auth_user in auth_users:
        try:
            plugin = get_authenticator_plugin(auth_user.provider.type)
            if plugin.type == 'local':
                return True
        except ImportError:
            pass

    return False



import logging
import time

from django.utils.translation import gettext_lazy as _
from rest_framework.views import APIView

from ansible_base.lib.utils.settings import get_function_from_setting, get_setting

logger = logging.getLogger('ansible_base.lib.utils.views.ansible_base')


class AnsibleBaseView(APIView):

    # pulp openapi generator compatibility
    endpoint_name = ''

    # pulp openapi generator compatibility
    @classmethod
    def endpoint_pieces(cls):
        return []

    def initialize_request(self, request, *args, **kwargs):
        """
        Store the Django REST Framework Request object as an attribute on the
        normal Django request, store time the request started.
        """
        self.time_started = time.time()

        return super().initialize_request(request, *args, **kwargs)

    def finalize_response(self, request, response, *args, **kwargs):
        response = super().finalize_response(request, response, *args, **kwargs)

        if request.user and request.user.is_authenticated:
            version = _('Unknown')
            setting = 'ANSIBLE_BASE_PRODUCT_VERSION_FUNCTION'
            the_function = None
            try:
                the_function = get_function_from_setting(setting)
            except Exception:
                logger.exception(_('Failed to load function from {setting} (see exception)'.format(setting=setting)))

            if the_function:
                try:
                    version = the_function()
                except Exception:
                    logger.exception(_('{setting} was set but calling it as a function failed (see exception).'.format(setting=setting)))

            response['X-API-Product-Version'] = version

        response['X-API-Product-Name'] = get_setting('ANSIBLE_BASE_PRODUCT_NAME', _('Unnamed'))
        response['X-API-Node'] = get_setting('CLUSTER_HOST_ID', _('Unknown'))

        time_started = getattr(self, 'time_started', None)
        if time_started:
            time_elapsed = time.time() - self.time_started
            response['X-API-Time'] = '%0.3fs' % time_elapsed

        if getattr(self, 'deprecated', False):
            response['Warning'] = _('This resource has been deprecated and will be removed in a future release.')

        return response

    def extra_related_fields(self, obj):
        """
        A hook for adding extra related fields to serializers which
        make use of this view/viewset.

        This is particularly useful for mixins which want to extend a viewset
        with additional actions and provide those actions as related fields.
        """
        return {}



import logging

from django.contrib.auth import BACKEND_SESSION_KEY
from django.core.exceptions import ImproperlyConfigured
from django.utils.deprecation import MiddlewareMixin
from social_django.middleware import SocialAuthExceptionMiddleware

from ansible_base.authentication.authenticator_plugins.utils import get_authenticator_plugins

logger = logging.getLogger('ansible_base.authentication.middleware')


def get_authenticator_module_paths() -> list:
    plugins = get_authenticator_plugins()
    plugins = [f'{name}.AuthenticatorPlugin' for name in plugins]
    return plugins


class AuthenticatorBackendMiddleware(MiddlewareMixin):
    _plugins = None

    @property
    def plugins(self):
        if not self._plugins:
            self._plugins = get_authenticator_module_paths()
        return self._plugins

    def process_request(self, request):
        if not hasattr(request, "session"):
            raise ImproperlyConfigured(
                "The Django AuthenticatorBackendMiddleware requires session "
                "middleware to be installed. Edit your MIDDLEWARE setting to "
                "insert "
                "'django.contrib.sessions.middleware.SessionMiddleware' before "
                "'AuthenticatorBackendMiddleware'."
            )

        # If the session backend is one from one of the Authenticator plugins, change it to
        # ansible_base.authentication.backend.AnsibleBaseAuth so that the user can be logged
        # in since the Authenticator backends aren't in AUTHENTICATION_BACKENDS
        if backend := request.session.get(BACKEND_SESSION_KEY, None):
            if backend in self.plugins:
                request.session[BACKEND_SESSION_KEY] = "ansible_base.authentication.backend.AnsibleBaseAuth"


class SocialExceptionHandlerMiddleware(SocialAuthExceptionMiddleware):
    def get_redirect_uri(self, request, exception):
        strategy = getattr(request, "social_strategy", None)
        error_url = strategy.setting("LOGIN_ERROR_URL")
        backend = getattr(request, "backend", None)
        backend_name = getattr(backend, "name", "unknown-backend")
        logger.error(f"Auth failure for backend {backend_name} - {repr(exception)}, redirecting to {error_url}")
        return error_url



try:
    from tabulate import tabulate

    HAS_TABULATE = True
except ImportError:
    HAS_TABULATE = False

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext_lazy as _

from ansible_base.authentication.models import Authenticator, AuthenticatorUser


class Command(BaseCommand):
    help = "Initialize service configuration with an admin user and a local authenticator"

    def add_arguments(self, parser):
        parser.add_argument("--list", action="store_true", help="list the authenticators", required=False)
        parser.add_argument("--initialize", action="store_true", help="Initialize an admin user and local db authenticator", required=False)
        parser.add_argument("--enable", type=int, help="Initialize an admin user and local db authenticator", required=False)
        parser.add_argument("--disable", type=int, help="Initialize an admin user and local db authenticator", required=False)

    def handle(self, *args, **options):
        took_action = False
        if options["initialize"]:
            self.initialize_authenticators()
            took_action = True
        if options["enable"] or options['disable']:
            for id, state in [(options['enable'], True), (options['disable'], False)]:
                if not id:
                    continue
                try:
                    authenticator = Authenticator.objects.get(id=id)
                except Authenticator.DoesNotExist:
                    raise CommandError(_("Authenticator %(id)s does not exist") % {"id": id})
                if authenticator.enabled is not state:
                    authenticator.enabled = state
                    authenticator.save()
            took_action = True
        if options["list"] or not took_action:
            self.list_authenticators()

    def list_authenticators(self):
        authenticators = []
        headers = ["ID", "Enabled", "Name", "Order"]

        for authenticator in Authenticator.objects.all().order_by('id'):
            authenticators.append([f'{authenticator.id}', f'{authenticator.enabled}', authenticator.name, f'{authenticator.order}'])

        self.stdout.write('')
        if HAS_TABULATE:
            self.stdout.write(tabulate(authenticators, headers, tablefmt="github"))
        else:
            self.stdout.write("\t".join(headers))
            for authenticator_data in authenticators:
                self.stdout.write("\t".join(authenticator_data))
        self.stdout.write('')

    def initialize_authenticators(self):
        admin_user = get_user_model().objects.filter(username="admin").first()
        if not admin_user:
            raise CommandError("Admin user with username 'admin' not defined.")

        existing_authenticator = Authenticator.objects.filter(type="ansible_base.authentication.authenticator_plugins.local").first()
        if not existing_authenticator:
            existing_authenticator = Authenticator.objects.create(
                name='Local Database Authenticator',
                enabled=True,
                create_objects=True,
                configuration={},
                created_by=admin_user,
                modified_by=admin_user,
                remove_users=False,
                type='ansible_base.authentication.authenticator_plugins.local',
            )
            self.stdout.write("Created default local authenticator")

            AuthenticatorUser.objects.get_or_create(
                uid=admin_user.username,
                user=admin_user,
                provider=existing_authenticator,
            )



from functools import partial

from django.apps import AppConfig
from django.db.models.signals import m2m_changed, post_save, pre_delete, pre_save

from ansible_base.activitystream.signals import activitystream_create, activitystream_delete, activitystream_m2m_changed, activitystream_update


def connect_activitystream_signals(cls):
    post_save.connect(activitystream_create, sender=cls, dispatch_uid=f'dab_activitystream_{cls.__name__}_create')
    pre_save.connect(activitystream_update, sender=cls, dispatch_uid=f'dab_activitystream_{cls.__name__}_update')
    pre_delete.connect(activitystream_delete, sender=cls, dispatch_uid=f'dab_activitystream_{cls.__name__}_delete')

    # Connect to m2m_changed signal for all m2m fields
    for field in cls._meta.many_to_many:
        if field.name in cls.activity_stream_excluded_field_names:
            continue

        fn = partial(activitystream_m2m_changed, field_name=field.name)
        m2m_changed.connect(
            fn,
            sender=getattr(cls, field.name).through,
            dispatch_uid=f'dab_activitystream_{cls.__name__}_{field.name}_m2m_changed',
            weak=False,
        )


class ActivitystreamConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'ansible_base.activitystream'
    label = 'dab_activitystream'
    verbose_name = 'Activity Stream'

    def ready(self):
        # Iterate models and look for ones that inherit from AuditableModel.
        # If found, connect the signal to the model by calling its static connect_signals().
        from django.apps import apps

        from ansible_base.activitystream.models import AuditableModel

        for model in apps.get_models():
            if issubclass(model, AuditableModel):
                connect_activitystream_signals(model)



from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _
from oauth2_provider.generators import generate_client_secret

from ansible_base.lib.serializers.common import NamedCommonModelSerializer
from ansible_base.lib.utils.encryption import ENCRYPTED_STRING
from ansible_base.oauth2_provider.models import OAuth2Application


class OAuth2ApplicationSerializer(NamedCommonModelSerializer):
    oauth2_client_secret = None

    class Meta:
        model = OAuth2Application
        fields = NamedCommonModelSerializer.Meta.fields + [x.name for x in OAuth2Application._meta.concrete_fields]
        read_only_fields = ('client_id', 'client_secret')
        read_only_on_update_fields = ('user', 'authorization_grant_type')
        extra_kwargs = {
            'user': {'allow_null': True, 'required': False},
            'organization': {'allow_null': False},
            'authorization_grant_type': {'allow_null': False, 'label': _('Authorization Grant Type')},
            'client_secret': {'label': _('Client Secret')},
            'client_type': {'label': _('Client Type')},
            'redirect_uris': {'label': _('Redirect URIs')},
            'app_url': {'label': _('Application URL')},
            'skip_authorization': {'label': _('Skip Authorization')},
        }

    def _get_client_secret(self, obj):
        request = self.context.get('request', None)
        try:
            if obj.client_type == 'public':
                return None
            elif request.method == 'POST':
                # Show the secret, one time, on POST
                return self.oauth2_client_secret
            else:
                return ENCRYPTED_STRING
        except ObjectDoesNotExist:
            return ''

    def to_representation(self, instance):
        # We have to override this because in AbstractCommonModelSerializer, we'll
        # auto-force all encrypted fields to ENCRYPTED_STRING. Usually that's fine,
        # but we want to show the client_secret on POST. Ideally we'd just use
        # get_client_secret() and a SerializerMethodField.
        ret = super().to_representation(instance)
        secret = self._get_client_secret(instance)
        if secret is None:
            del ret['client_secret']
        else:
            ret['client_secret'] = secret
        return ret

    def _summary_field_tokens(self, obj):
        token_list = [{'id': x.pk, 'token': ENCRYPTED_STRING, 'scope': x.scope} for x in obj.access_tokens.all()[:10]]
        if len(token_list) < 10:
            token_count = len(token_list)
        else:
            token_count = obj.access_tokens.count()
        return {'count': token_count, 'results': token_list}

    def _get_summary_fields(self, obj) -> dict[str, dict]:
        ret = super()._get_summary_fields(obj)
        ret['tokens'] = self._summary_field_tokens(obj)
        return ret

    def create(self, validated_data):
        # This is hacky:
        # There is a cascading set of issues here.
        # 1. The first thing to know is that DOT automatically hashes the client_secret
        #    in a pre_save method on the client_secret field.
        # 2. In current released versions, there is no way to disable (1). It uses
        #    the built-in Django password hashing stuff to do this. There's a merged
        #    PR to allow disabling this (DOT #1311), but it's not released yet.
        # 3. If we use our own encrypted_field stuff, it conflicts with (1) and (2).
        #    They end up giving our encrypted field to Django's password check
        #    and *we* end up showing *their* hashed value to the user on POST, which
        #    doesn't work, the user needs to see the real (decrypted) value. So
        #    until upstream #1311 is released, we do NOT treat the field as an
        #    encrypted_field, we just defer to the upstream hashing.
        # 4. But we have no way to see the client_secret on POST, if we let the
        #    model generate it, because it's hashed by the time we get to the
        #    serializer...
        #
        # So to that end, on POST, we'll make the client secret here, and then
        # we can access it to show the user the value (once) on POST.
        validated_data['client_secret'] = generate_client_secret()
        self.oauth2_client_secret = validated_data['client_secret']
        return super().create(validated_data)



from contextlib import contextmanager
from zlib import crc32

from django.db import DEFAULT_DB_ALIAS, connection, connections, transaction
from django.db.migrations.executor import MigrationExecutor


@contextmanager
def ensure_transaction():
    needs_new_transaction = not transaction.get_connection().in_atomic_block

    if needs_new_transaction:
        with transaction.atomic():
            yield
    else:
        yield


def migrations_are_complete() -> bool:
    """Returns a boolean telling you if manage.py migrate has been run to completion

    Note that this is a little expensive, like up to 20 database queries
    and lots of imports.
    Not suitable to run as part of a request, but expected to be okay
    in a management command or post_migrate signals"""
    executor = MigrationExecutor(connection)
    plan = executor.migration_plan(executor.loader.graph.leaf_nodes())
    return not bool(plan)


# NOTE: the django_pglocks_advisory_lock context manager was forked from the django-pglocks v1.0.4
# that was licensed under the MIT license


@contextmanager
def django_pglocks_advisory_lock(lock_id, shared=False, wait=True, using=None):

    if using is None:
        using = DEFAULT_DB_ALIAS

    # Assemble the function name based on the options.

    function_name = 'pg_'

    if not wait:
        function_name += 'try_'

    function_name += 'advisory_lock'

    if shared:
        function_name += '_shared'

    release_function_name = 'pg_advisory_unlock'
    if shared:
        release_function_name += '_shared'

    # Format up the parameters.

    tuple_format = False

    if isinstance(
        lock_id,
        (
            list,
            tuple,
        ),
    ):
        if len(lock_id) != 2:
            raise ValueError("Tuples and lists as lock IDs must have exactly two entries.")

        if not isinstance(lock_id[0], int) or not isinstance(lock_id[1], int):
            raise ValueError("Both members of a tuple/list lock ID must be integers")

        tuple_format = True
    elif isinstance(lock_id, str):
        # Generates an id within postgres integer range (-2^31 to 2^31 - 1).
        # crc32 generates an unsigned integer in Py3, we convert it into
        # a signed integer using 2's complement (this is a noop in Py2)
        pos = crc32(lock_id.encode("utf-8"))
        lock_id = (2**31 - 1) & pos
        if pos & 2**31:
            lock_id -= 2**31
    elif not isinstance(lock_id, int):
        raise ValueError("Cannot use %s as a lock id" % lock_id)

    if tuple_format:
        base = "SELECT %s(%d, %d)"
        params = (
            lock_id[0],
            lock_id[1],
        )
    else:
        base = "SELECT %s(%d)"
        params = (lock_id,)

    acquire_params = (function_name,) + params

    command = base % acquire_params
    cursor = connections[using].cursor()

    cursor.execute(command)

    if not wait:
        acquired = cursor.fetchone()[0]
    else:
        acquired = True

    try:
        yield acquired
    finally:
        if acquired:
            release_params = (release_function_name,) + params

            command = base % release_params
            cursor.execute(command)

        cursor.close()


@contextmanager
def advisory_lock(*args, lock_session_timeout_milliseconds=0, **kwargs):
    """Context manager that wraps the pglocks advisory lock

    This obtains a named lock in postgres, idenfied by the args passed in
    usually the lock identifier is a simple string.

    @param: wait If True, block until the lock is obtained
    @param: shared Whether or not the lock is shared
    @param: lock_session_timeout_milliseconds Postgres-level timeout
    @param: using django database identifier
    """
    if connection.vendor == "postgresql":
        cur = None
        idle_in_transaction_session_timeout = None
        idle_session_timeout = None
        if lock_session_timeout_milliseconds > 0:
            with connection.cursor() as cur:
                idle_in_transaction_session_timeout = cur.execute("SHOW idle_in_transaction_session_timeout").fetchone()[0]
                idle_session_timeout = cur.execute("SHOW idle_session_timeout").fetchone()[0]
                cur.execute("SET idle_in_transaction_session_timeout = %s", (lock_session_timeout_milliseconds,))
                cur.execute("SET idle_session_timeout = %s", (lock_session_timeout_milliseconds,))
        with django_pglocks_advisory_lock(*args, **kwargs) as internal_lock:
            yield internal_lock
            if lock_session_timeout_milliseconds > 0:
                with connection.cursor() as cur:
                    cur.execute("SET idle_in_transaction_session_timeout = %s", (idle_in_transaction_session_timeout,))
                    cur.execute("SET idle_session_timeout = %s", (idle_session_timeout,))
    elif connection.vendor == "sqlite":
        yield True
    else:
        raise RuntimeError(f'Advisory lock not implemented for database type {connection.vendor}')



from django.urls import include, path, re_path
from oauth2_provider import views as oauth_views

from ansible_base.lib.routers import AssociationResourceRouter
from ansible_base.oauth2_provider import views as oauth2_provider_views
from ansible_base.oauth2_provider.apps import Oauth2ProviderConfig

app_name = Oauth2ProviderConfig.label

router = AssociationResourceRouter()

router.register(
    r'applications',
    oauth2_provider_views.OAuth2ApplicationViewSet,
    basename='application',
    related_views={
        'tokens': (oauth2_provider_views.OAuth2TokenViewSet, 'access_tokens'),
    },
)

router.register(
    r'tokens',
    oauth2_provider_views.OAuth2TokenViewSet,
    basename='token',
)

api_version_urls = [
    path('', include(router.urls)),
]

root_urls = [
    re_path(r'^o/$', oauth2_provider_views.ApiOAuthAuthorizationRootView.as_view(), name='oauth_authorization_root_view'),
    re_path(r"^o/authorize/$", oauth_views.AuthorizationView.as_view(), name="authorize"),
    re_path(r"^o/token/$", oauth2_provider_views.TokenView.as_view(), name="token"),
    re_path(r"^o/revoke_token/$", oauth_views.RevokeTokenView.as_view(), name="revoke-token"),
]



import copy
import logging
from itertools import chain
from typing import Type

from django.conf import settings
from django.db.models import Model
from django.db.models.fields import IntegerField
from django.db.models.query import QuerySet
from django.http import QueryDict
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from rest_framework import routers, serializers, status
from rest_framework.decorators import action
from rest_framework.permissions import SAFE_METHODS
from rest_framework.request import clone_request
from rest_framework.response import Response
from rest_framework.viewsets import ViewSetMixin

from ansible_base.rbac.permission_registry import permission_registry

logger = logging.getLogger('ansible_base.lib.routers.association_resource_router')


# Registry contains subclasses of AssociationSerializerBase indexed by name
# this prevents duplicate names which would cause schema to not render correctly
serializer_registry = {}


class AssociationSerializerBase(serializers.Serializer):
    """It is expected that final subclasses will set the instances field"""

    instances = None

    def get_queryset_on_init(self, original_qs: QuerySet) -> QuerySet:
        raise NotImplementedError

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        new_qs = self.get_queryset_on_init(self.fields['instances'].child_relation.queryset)
        self.fields['instances'].child_relation.queryset = new_qs


class DisassociationSerializerBase(AssociationSerializerBase):
    """Serializer used for removing objects that are currently associated via a many-to-many relationship"""

    def get_queryset_on_init(self, original_qs: QuerySet) -> QuerySet:
        if 'view' in self.context:
            view = self.context['view']
            if 'pk' in view.kwargs:
                return view.get_queryset()
        return original_qs

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['instances'].child_relation.error_messages = self.fields['instances'].child_relation.error_messages.copy()
        self.fields['instances'].child_relation.error_messages['does_not_exist'] = _(
            'Invalid pk "{pk_value}" - object does not exist or is not associated with parent object.'
        )


class FilteredAssociationSerializer(AssociationSerializerBase):
    """Serializer used for adding objects to a many-to-many relationship"""

    def get_queryset_on_init(self, original_qs: QuerySet) -> QuerySet:
        if 'view' in self.context:
            # If the view exists we require it to be an instance of AssociationViewSetMethodsMixin
            view = self.context['view']
            return view.filter_associate_queryset(original_qs)
        return original_qs


class AssociationViewSetMethodsMixin:
    """Contains methods called by viewset methods for list, associate, disassociate actions

    Importantly, this is placed high up in the inheritance chain.
    This means that the viewset passed in for the related views can override these methods.
    General principal is that these should not overlap with DRF methods, because those
    are likely already defined by the passed-in viewset and will not take effect.
    """

    def filter_associate_queryset(self, qs: QuerySet) -> QuerySet:
        """Limits queryset to objects allowed to be associated

        This is allowed to be different from the filter for the list view,
        which affects what is visible to the user.
        Compare to this, which defines what objects can be _associated_
        probably due to permission restrictions... normally based on visibility.
        """
        return self.filter_queryset(qs)

    def get_sublist_queryset(self, parent_instance: Model) -> QuerySet:
        """Queryset to show, given a parent object

        Example would be /organizations/N/teams/
        This method returns the team queryset, given organization object pk=N
        """
        return getattr(parent_instance, self.association_fk).all()

    def perform_associate(self, parent_instance: Model, related_instances: list[Model]) -> None:
        """Attach related_instances to instance via the relationship this viewset manages"""
        manager = getattr(parent_instance, self.association_fk)
        manager.add(*related_instances)

    def perform_disassociate(self, parent_instance: Model, related_instances: list[Model]) -> None:
        """Remove related_instances from the managed relationship of instance"""
        manager = getattr(parent_instance, self.association_fk)
        manager.remove(*related_instances)


class RelatedListMixin:
    """Mixin used for related viewsets which contain a sub-list, like /organizations/N/teams/"""

    def check_parent_object_permissions(self, request, parent_obj: Model) -> None:
        """Check that request user has permission to parent_obj

        Associate and disassociate is a POST request, list is GET
        the normal process of get_object --> check_object_permissions
        will not check "change" permissions to the parent object on POST
        this method checks parent change permission, view permission should be handled by filter_queryset
        """
        if (request.method not in SAFE_METHODS) and 'ansible_base.rbac' in settings.INSTALLED_APPS and permission_registry.is_registered(parent_obj):
            from ansible_base.rbac.policies import check_content_obj_permission

            check_content_obj_permission(request.user, parent_obj)

    def get_parent_object(self) -> Model:
        """Modeled mostly after DRF get_object, but for the parent model

        Like for /api/v2/organizations/<pk>/cows/, this returns the organization
        with the specified pk.
        """
        parent_view = self.parent_viewset()
        parent_view.request = clone_request(self.request, 'GET')
        parent_view.request._request = copy.copy(self.request._request)
        parent_view.request._request.GET = QueryDict()
        queryset = parent_view.filter_queryset(parent_view.get_queryset())
        filter_kwargs = {'pk': self.kwargs['pk']}
        parent_obj = get_object_or_404(queryset, **filter_kwargs)

        # May raise a permission denied
        self.check_parent_object_permissions(self.request, parent_obj)

        return parent_obj

    def get_queryset(self) -> QuerySet:
        parent_instance = self.get_parent_object()
        return self.get_sublist_queryset(parent_instance)


class AssociateMixin(RelatedListMixin):
    """Mixin used for writable related viewsets, where objects can be associated or disassociated from the relationship"""

    instances_help_text = {
        'associate': _('A list of {model_name} to add to this relationship.'),
        'disassociate': _('A list of {model_name} to remove from this relationship.'),
    }
    parent_serializer_cls = {'associate': FilteredAssociationSerializer, 'disassociate': DisassociationSerializerBase}

    @action(detail=False, methods=['post'])
    def associate(self, request, **kwargs):
        instance = self.get_parent_object()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        related_instances = serializer.validated_data['instances']
        if not related_instances:
            raise serializers.ValidationError({'instances': _('Please pass in one or more instances to associate')})

        self.perform_associate(instance, related_instances)

        return Response(status=status.HTTP_204_NO_CONTENT)

    @action(detail=False, methods=['post'])
    def disassociate(self, request, **kwargs):
        instance = self.get_parent_object()
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        related_instances = serializer.validated_data['instances']
        if not related_instances:
            raise serializers.ValidationError({'instances': _('Please pass in one or more instances to disassociate')})

        self.perform_disassociate(instance, related_instances)

        return Response(status=status.HTTP_204_NO_CONTENT)

    def get_viewset_model(self) -> Type[Model]:
        """Return model for this viewset"""
        if self.queryset:
            return self.queryset.model
        return self.serializer_class.Meta.model

    def get_serializer_class(self) -> Type[serializers.BaseSerializer]:
        if self.action in ('disassociate', 'associate'):
            cls = self.get_viewset_model()
            pretty_model_name = cls._meta.verbose_name.title().replace(' ', '')
            cls_name = f'{pretty_model_name}{self.action.title()}Serializer'

            default_instances_field = serializers.PrimaryKeyRelatedField(
                queryset=cls.objects.all(), many=True, help_text=self.instances_help_text[self.action].format(model_name=cls._meta.verbose_name_plural)
            )

            if cls_name not in serializer_registry:
                serializer_registry[cls_name] = type(cls_name, (self.parent_serializer_cls[self.action],), {'instances': default_instances_field})
            return serializer_registry[cls_name]

        return super().get_serializer_class()


@property
def attribute_raiser(cls):
    """This method is used to exclude methods in subclass of provided viewset

    How it works:
    SimpleRouter.get_method_map determines what actions are in a viewset with hasattr.
    raising AttributeError makes that return False, so methods like retrieve
    will not be included in the returned method map, so it will not build a URL
    for those, because we do not want them in the association URL set.
    """
    raise AttributeError


class AssociationResourceRouter(routers.SimpleRouter):
    docstring_template = _(
        """Manage {child_model} objects in the '{relationship}' relationship of this particular {parent_model}.

        Starting from the detail URL:

        GET /:id/{relationship}/ to show {child_model_plural} currently in the relationship"""
    )
    reverse_docstring = _(
        """

        POST a list of instances to /:id/{relationship}/associate/ to add those {child_model_plural} to the relationship

        POST a list of instances to /:id/{relationship}/disassociate/ to remove those {child_model_plural} from the relationshp
        """
    )

    def get_method_map(self, viewset, method_map):
        is_associate_viewset = issubclass(viewset, AssociateMixin)
        associate_actions = ['associate', 'disassociate', 'list']
        bound_methods = {}
        for method, action_str in method_map.items():
            if hasattr(viewset, action_str):
                if is_associate_viewset and action_str not in associate_actions:
                    continue
                bound_methods[method] = action_str
        return bound_methods

    def associated_viewset_cls_factory(self, viewset: Type[ViewSetMixin]) -> Type[ViewSetMixin]:
        """Given viewset (as a class) return a subclass containing all its actions except for list"""

        custom_methods = chain.from_iterable(action.mapping.values() for action in viewset.get_extra_actions())

        exclude_list = ('retrieve', 'update', 'partial_update', 'destroy') + tuple(custom_methods)

        class AssociatedViewSetType(type(viewset)):
            """Metaclass that turns off viewset methods other than list"""

            method_excludes = exclude_list

            def __dir__(self):
                ret = list(super().__dir__())
                for method in self.method_excludes:
                    if method in ret:
                        ret.remove(method)
                return ret

        for method in exclude_list:
            setattr(AssociatedViewSetType, method, attribute_raiser)

        class AssociatedViewSet(viewset, AssociationViewSetMethodsMixin, metaclass=AssociatedViewSetType):
            """Adjusted version of given viewset for related endpoint with only list views"""

            pass

        return AssociatedViewSet

    def register(self, prefix, viewset, related_views={}, basename=None):
        if basename is None:
            basename = self.get_default_basename(viewset)

        for related_name, (related_view, fk) in related_views.items():
            parent_model = viewset.serializer_class.Meta.model
            child_model = related_view.serializer_class.Meta.model

            # Determine if this is a related view or a reverse view
            is_reverse_view = bool(any(x.related_model == child_model for x in parent_model._meta.related_objects))
            mixin_class = AssociateMixin
            docstring_template = self.docstring_template
            if is_reverse_view:
                mixin_class = RelatedListMixin
            else:
                docstring_template += self.reverse_docstring

            # Start with a viewset that only has list action enabled
            associated_viewset = self.associated_viewset_cls_factory(related_view)

            docstring_kwargs = {
                'child_model': child_model._meta.verbose_name.title(),
                'parent_model': parent_model._meta.verbose_name.title(),
                'child_model_plural': child_model._meta.verbose_name_plural,
                'parent_model_plural': parent_model._meta.verbose_name_plural,
                'relationship': fk,
            }

            # Generate the related viewset
            # Name includes and parent and child viewset, because this defines global uniqueness
            modified_related_viewset = type(
                f'Related{viewset.__name__}{related_view.__name__}',
                (mixin_class, associated_viewset),
                {
                    '__doc__': docstring_template.format(**docstring_kwargs),
                    'association_fk': fk,
                    'parent_viewset': viewset,
                    'lookup_field': fk,
                },
            )

            # Force a reverse view to be read only
            if is_reverse_view:
                modified_related_viewset.http_method_names = ['get', 'head', 'options']

            if isinstance(child_model._meta.pk, IntegerField):
                url_path = f"{prefix}/(?P<pk>[0-9]+)/{related_name}"
            else:
                url_path = f"{prefix}/(?P<pk>[^/.]+)/{related_name}"

            # Register the viewset
            self.registry.append((url_path, modified_related_viewset, f'{basename}-{fk}'))

        super().register(prefix, viewset, basename)



import logging
from urllib.parse import urljoin, urlparse

import requests
from django.utils.translation import gettext as _

from ansible_base.jwt_consumer.common.cache import JWTCache
from ansible_base.lib.utils.settings import get_setting

logger = logging.getLogger('ansible_base.jwt_consumer.common.cert')


class JWTCertException(Exception):
    '''
    Raised if we can't get the cert for any reason.
    The message will already be translated.
    '''

    pass


class JWTCert:
    key_name = 'ANSIBLE_BASE_JWT_KEY'

    def __init__(self):
        self.cached = None
        self.key = None
        # Attempt to locate the cert using ANSIBLE_BASE_JWT_KEY.  If we are running on a service that houses the JWT key
        #  we should not have that setting set and instead should have that setting in jwt_public_key so fallback to that
        self.jwt_key_setting = get_setting(self.key_name, get_setting('jwt_public_key', None))
        self.cache = JWTCache()

    def _get_decryption_key_from_url(self) -> None:
        url = self.jwt_key_setting
        validate_certs = get_setting("ANSIBLE_BASE_JWT_VALIDATE_CERT", True)
        timeout = get_setting("ANSIBLE_BASE_JWT_URL_TIMEOUT", 30)

        # If the URL does not end with / the urljoin will wipe out the existing path
        if not url.endswith('/'):
            url = f"{url}/"
        jwt_key_url = urljoin(url, "api/gateway/v1/jwt_key/")

        logger.debug(f"Loading decryption key from url {jwt_key_url}")

        try:
            response = requests.get(
                jwt_key_url,
                verify=validate_certs,
                timeout=timeout,
            )
            if response.status_code != 200:
                raise JWTCertException(_("Failed to get 200 response from the issuer: {0}").format(response.status_code))
            self.key = response.text
        except requests.exceptions.ConnectionError as e:
            raise JWTCertException(_("Failed to connect to {0}: {1}").format(jwt_key_url, e))
        except requests.exceptions.Timeout:
            raise JWTCertException(_("Timed out after {0} secs when connecting to {1}").format(timeout, jwt_key_url))
        except requests.exceptions.RequestException as e:
            raise JWTCertException(_("Failed to get JWT decryption key from JWT server: ({0}) {1}").format(e.__class__.__name__, e))

    def _get_decryption_key_from_file(self, file_path: str) -> None:
        logger.debug(f"Loading decryption key from file {file_path}")

        try:
            with open(file_path, "r") as f:
                self.key = f.read()
        except FileNotFoundError:
            raise JWTCertException(_("The specified file {0} does not exist").format(file_path))
        except IsADirectoryError:
            raise JWTCertException(_("The specified file {0} is not a file").format(file_path))
        except PermissionError:
            raise JWTCertException(_("Permission error when reading {0}").format(file_path))
        except Exception as e:
            raise JWTCertException(_("Failed reading {0}: {1}").format(file_path, e))

    def get_decryption_key(self, ignore_cache: bool = False) -> None:
        # Set key and cached to None
        self.key = None
        self.cached = None

        if self.jwt_key_setting is None:
            logger.info(f"Failed to get the setting {self.key_name}")
            return

        cached_key = self.cache.get_key_from_cache()
        if cached_key and not ignore_cache:
            logger.debug(f"Loading decryption key from cache instead of from url {self.jwt_key_setting}")
            self.cached = True
            self.key = cached_key
            return

        # We don't check the cache right away here because we only want to check the cache if its a file or URL.
        # A hard coded key would be less efficient if we were attempting to load the cache every time.
        url_or_string = self.jwt_key_setting
        url_info = urlparse(url_or_string)
        logger.info(f"Loading decryption key from {url_or_string} scheme {url_info.scheme}")
        if url_info.scheme in ["http", "https"]:
            self._get_decryption_key_from_url()
        elif url_info.scheme == "file":
            file_path = url_info.path
            self._get_decryption_key_from_file(file_path)
        elif url_info.scheme == "" and url_info.path != "":
            logger.debug("Assuming decryption key is the actual cert")
            self.key = url_or_string

        if self.key is None:
            raise JWTCertException(_("Unable to determine how to handle {0} to get key").format(url_or_string))
        elif not self.key.startswith('-----BEGIN PUBLIC KEY-----') and not self.key.endswith('-----END PUBLIC KEY-----'):
            logger.debug(self.key)
            raise JWTCertException(_("Returned key does not start and end with BEGIN/END PUBLIC KEY"))
        logger.info("Decryption key appears valid")
        logger.debug(f"{self.key}")
        self.cache.set_key_in_cache(self.key)
        self.cached = False



# Generated by Django 4.2.6 on 2023-11-20 20:48

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        migrations.swappable_dependency(settings.ANSIBLE_BASE_TEAM_MODEL),
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='DABPermission',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='name')),
                ('codename', models.CharField(max_length=100, verbose_name='codename')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='content type')),
            ],
            options={
                'verbose_name': 'permission',
                'verbose_name_plural': 'permissions',
                'ordering': ['content_type__model', 'codename'],
                'unique_together': {('content_type', 'codename')},
            },
        ),
        migrations.CreateModel(
            name='RoleDefinition',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField(db_index=True, unique=True)),
                ('description', models.TextField(blank=True)),
                ('managed', models.BooleanField(default=False, editable=False)),
                ('permissions', models.ManyToManyField(related_name='role_definitions', to='dab_rbac.DABPermission')),
                ('content_type', models.ForeignKey(
                    default=None,
                    help_text='Type of resource this can apply to, only used for validation and user assistance',
                    null=True,
                    on_delete=django.db.models.deletion.CASCADE,
                    to='contenttypes.contenttype'
                )),
                ('created_by', models.ForeignKey(
                    default=None,
                    editable=False,
                    help_text='The user who created this resource',
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name='%(app_label)s_%(class)s_created+',
                    to=settings.AUTH_USER_MODEL
                )),
                ('created', models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created')),
                ('modified_by', models.ForeignKey(
                    default=None,
                    editable=False,
                    help_text='The user who last modified this resource',
                    null=True,
                    on_delete=django.db.models.deletion.SET_NULL,
                    related_name='%(app_label)s_%(class)s_modified+',
                    to=settings.AUTH_USER_MODEL
                )),
                ('modified', models.DateTimeField(auto_now=True, help_text='The date/time this resource was created')),
            ],
            options={
                'ordering': ['id'],
                'verbose_name_plural': 'role_definition',
            },
        ),
        migrations.CreateModel(
            name='RoleTeamAssignment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('object_id', models.TextField(
                    blank=True,
                    help_text='Primary key of the object this assignment applies to, null value indicates system-wide assignment',
                    null=True)),
                ('role_definition', models.ForeignKey(
                    help_text='The role definition which defines permissions conveyed by this assignment', on_delete=django.db.models.deletion.CASCADE,
                    related_name='team_assignments', to='dab_rbac.roledefinition')),
            ],
            options={
                'ordering': ['id'],
            },
        ),
        migrations.AddField(
            model_name='roleteamassignment',
            name='created_by',
            field=models.ForeignKey(
                default=None,
                editable=False,
                help_text='The user who created this resource',
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='%(app_label)s_%(class)s_created+',
                to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name='roleteamassignment',
            name='team',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.ANSIBLE_BASE_TEAM_MODEL, related_name='role_assignments'),
        ),
        migrations.CreateModel(
            name='RoleUserAssignment',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='The date/time this resource was created')),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('object_id', models.TextField(
                    blank=True,
                    help_text='Primary key of the object this assignment applies to, null value indicates system-wide assignment',
                    null=True)),
                ('role_definition', models.ForeignKey(
                    help_text='The role definition which defines permissions conveyed by this assignment', on_delete=django.db.models.deletion.CASCADE,
                    related_name='user_assignments', to='dab_rbac.roledefinition')),
            ],
            options={
                'ordering': ['id'],
            },
        ),
        migrations.AddField(
            model_name='roleuserassignment',
            name='created_by',
            field=models.ForeignKey(
                default=None,
                editable=False,
                help_text='The user who created this resource',
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='%(app_label)s_%(class)s_created+',
                to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name='roleuserassignment',
            name='user',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, related_name='role_assignments'),
        ),
        migrations.CreateModel(
            name='ObjectRole',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.TextField(null=False)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('provides_teams', models.ManyToManyField(
                    help_text='Users who have this role obtain member access to these teams, and inherit all their permissions',
                    related_name='member_roles', to=settings.ANSIBLE_BASE_TEAM_MODEL)),
                ('role_definition', models.ForeignKey(
                    help_text='The role definition which defines what permissions this object role grants', on_delete=django.db.models.deletion.CASCADE,
                    related_name='object_roles', to='dab_rbac.roledefinition')),
            ],
            options={
                'verbose_name_plural': 'object_roles',
                'ordering': ('content_type', 'object_id'),
            },
        ),
        migrations.AddField(
            model_name='roleuserassignment',
            name='object_role',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='dab_rbac.objectrole'),
        ),
        migrations.AddField(
            model_name='roleteamassignment',
            name='object_role',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='dab_rbac.objectrole'),
        ),
        migrations.AddField(
            model_name='objectrole',
            name='teams',
            field=models.ManyToManyField(
                help_text='Teams or groups who have access to the permissions defined by this object role',
                related_name='has_roles',
                through='dab_rbac.RoleTeamAssignment',
                through_fields=("object_role", "team"),
                to=settings.ANSIBLE_BASE_TEAM_MODEL
            ),
        ),
        migrations.AddField(
            model_name='objectrole',
            name='users',
            field=models.ManyToManyField(
                help_text='Users who have access to the permissions defined by this object role',
                related_name='has_roles',
                through='dab_rbac.RoleUserAssignment',
                through_fields=("object_role", "user"),
                to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.CreateModel(
            name='RoleEvaluation',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('codename', models.TextField(help_text='The name of the permission, giving the action and the model, from the Django Permission model')),
                ('content_type_id', models.PositiveIntegerField()),
                ('object_id', models.PositiveIntegerField(null=False)),
            ],
            options={
                'verbose_name_plural': 'role_object_permissions',
            },
        ),
        migrations.AddIndex(
            model_name='objectrole',
            index=models.Index(fields=['content_type', 'object_id'], name='dab_rbac_ob_content_cbd55d_idx'),
        ),
        migrations.AddConstraint(
            model_name='objectrole',
            constraint=models.UniqueConstraint(fields=('object_id', 'content_type', 'role_definition'), name='one_object_role_per_object_and_role'),
        ),
        migrations.AddField(
            model_name='roleevaluation',
            name='role',
            field=models.ForeignKey(
                help_text='The object role that grants this form of permission',
                on_delete=django.db.models.deletion.CASCADE,
                related_name='permission_partials',
                to='dab_rbac.objectrole'
            ),
        ),
        migrations.AddIndex(
            model_name='roleevaluation',
            index=models.Index(fields=['role', 'content_type_id', 'object_id'], name='dab_rbac_ro_role_id_604bc4_idx'),
        ),
        migrations.AddIndex(
            model_name='roleevaluation',
            index=models.Index(fields=['role', 'content_type_id', 'codename'], name='dab_rbac_ro_role_id_8b9faf_idx'),
        ),
        migrations.AddConstraint(
            model_name='roleevaluation',
            constraint=models.UniqueConstraint(fields=('object_id', 'content_type_id', 'codename', 'role'), name='one_entry_per_object_permission_and_role'),
        ),
        migrations.AlterField(
            model_name='objectrole',
            name='provides_teams',
            field=models.ManyToManyField(
                editable=False,
                help_text='Users who have this role obtain member access to these teams, and inherit all their permissions',
                related_name='member_roles',
                to=settings.ANSIBLE_BASE_TEAM_MODEL
            ),
        ),
        migrations.CreateModel(
            name='RoleEvaluationUUID',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('codename', models.TextField(help_text='The name of the permission, giving the action and the model, from the Django Permission model')),
                ('content_type_id', models.PositiveIntegerField()),
                ('object_id', models.UUIDField()),
                ('role', models.ForeignKey(
                    help_text='The object role that grants this form of permission',
                    on_delete=django.db.models.deletion.CASCADE,
                    related_name='permission_partials_uuid',
                    to='dab_rbac.objectrole'
                )),
            ],
            options={
                'verbose_name_plural': 'role_object_permissions',
                'indexes': [
                    models.Index(fields=['role', 'content_type_id', 'object_id'], name='dab_rbac_ro_role_id_237936_idx'),
                    models.Index(fields=['role', 'content_type_id', 'codename'], name='dab_rbac_ro_role_id_4fe905_idx')
                ],
            },
        ),
        migrations.AddConstraint(
            model_name='roleevaluationuuid',
            constraint=models.UniqueConstraint(
                fields=('object_id', 'content_type_id', 'codename', 'role'), name='one_entry_per_object_permission_and_role_uuid'),
        ),
        migrations.AlterUniqueTogether(
            name='roleteamassignment',
            unique_together={('team', 'object_role')},
        ),
        migrations.AlterUniqueTogether(
            name='roleuserassignment',
            unique_together={('user', 'object_role')},
        ),
    ]



import logging

from django.db.models.fields import NOT_PROVIDED
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from rest_framework.exceptions import PermissionDenied
from rest_framework.fields import empty
from rest_framework.serializers import ValidationError

from ansible_base.lib.abstract_models.common import get_url_for_object
from ansible_base.lib.serializers.validation import ValidationSerializerMixin
from ansible_base.lib.utils import models
from ansible_base.lib.utils.encryption import ENCRYPTED_STRING

logger = logging.getLogger('ansible_base.lib.serializers.common')


class AbstractCommonModelSerializer(ValidationSerializerMixin, serializers.ModelSerializer):
    url = serializers.SerializerMethodField()
    related = serializers.SerializerMethodField('_get_related')
    summary_fields = serializers.SerializerMethodField('_get_summary_fields')

    class Meta:
        fields = ['id', 'url', 'related', 'summary_fields']

    def __init__(self, instance=None, data=empty, **kwargs):
        # pre-populate the form with the defaults from the model
        model = getattr(self.Meta, 'model', None)
        if model:
            extra_kwargs = getattr(self.Meta, 'extra_kwargs', {})
            for field in model._meta.concrete_fields:
                if field.name not in extra_kwargs:
                    extra_kwargs[field.name] = {}
                if not extra_kwargs[field.name].get('initial', None):
                    if field.default and field.default is not NOT_PROVIDED:
                        extra_kwargs[field.name]['initial'] = field.default
            setattr(self.Meta, 'extra_kwargs', extra_kwargs)
        super().__init__(instance, data, **kwargs)

    def get_url(self, obj) -> str:
        return get_url_for_object(obj)

    @property
    def is_list_view(self) -> bool:
        view = self.context.get('view')
        if view is None:
            raise ValueError("View not found in context")
        return view.action == 'list'

    # Type hints are used by OpenAPI
    def _get_related(self, obj) -> dict[str, str]:
        if obj is None:
            return {}
        related_fields = {}
        view = self.context.get('view')
        if view is not None and hasattr(view, 'extra_related_fields'):
            related_fields.update(view.extra_related_fields(obj))
        if not hasattr(obj, 'related_fields'):
            logger.warning(f"Object {obj.__class__} has no related_fields method")
        else:
            related_fields.update(obj.related_fields(self.context.get('request')))

        return related_fields

    def _get_summary_fields(self, obj) -> dict[str, dict]:
        if obj is None:
            return {}
        if not hasattr(obj, 'get_summary_fields'):
            logger.warning(f"Object {obj.__class__} has no get_summary_fields method")
            return {}
        return obj.get_summary_fields()

    def to_representation(self, obj):
        ret = super().to_representation(obj)

        for key in obj.encrypted_fields:
            if key in ret:
                ret[key] = ENCRYPTED_STRING

        return ret

    def update(self, instance, validated_data):
        # We don't want the $encrypted$ fields going back to the model
        for key in self.Meta.model.encrypted_fields:
            new_field = validated_data.get(key, None)
            if new_field and new_field == ENCRYPTED_STRING:
                validated_data.pop(key, None)

        return super().update(instance, validated_data)


class CommonModelSerializer(AbstractCommonModelSerializer):
    class Meta(AbstractCommonModelSerializer.Meta):
        fields = AbstractCommonModelSerializer.Meta.fields + ['created', 'created_by', 'modified', 'modified_by']


class NamedCommonModelSerializer(CommonModelSerializer):
    class Meta(CommonModelSerializer.Meta):
        fields = CommonModelSerializer.Meta.fields + ['name']


class ImmutableCommonModelSerializer(AbstractCommonModelSerializer):
    class Meta(AbstractCommonModelSerializer.Meta):
        fields = AbstractCommonModelSerializer.Meta.fields + ['created', 'created_by']


class CommonUserSerializer(CommonModelSerializer):
    """
    Disallows editing of system user and enforces superuser requirement.
    """

    def validate(self, data):
        if models.get_system_user() is None:
            return data
        if hasattr(self, 'instance') and hasattr(self.instance, 'id') and self.instance.id == models.get_system_user().id:
            raise ValidationError(_('System users cannot be modified'))
        return data

    def validate_is_superuser(self, value):
        if value is True:
            if not self.context['request'].user.is_superuser:
                raise PermissionDenied
        return value



from rest_framework.exceptions import APIException

# 498 is not a standard error code, so we have to manually define it
HTTP_498_INVALID_TOKEN = 498


class InvalidService(Exception):
    def __init__(self, service):
        super().__init__(f"This authentication class requires {service}.")


class InvalidTokenException(APIException):
    status_code = HTTP_498_INVALID_TOKEN
    status_text = "Invalid Token"
    default_detail = "Invalid or expired token."
    default_code = "invalid_token"



import hashlib
from datetime import timedelta

from django.utils.timezone import now
from oauth2_provider import views as oauth_views
from oauthlib import oauth2
from rest_framework.viewsets import ModelViewSet

from ansible_base.lib.utils.hashing import hash_string
from ansible_base.lib.utils.settings import get_setting
from ansible_base.lib.utils.views.django_app_api import AnsibleBaseDjangoAppApiView
from ansible_base.oauth2_provider.models import OAuth2AccessToken, OAuth2RefreshToken
from ansible_base.oauth2_provider.permissions import OAuth2ScopePermission
from ansible_base.oauth2_provider.serializers import OAuth2TokenSerializer
from ansible_base.oauth2_provider.views.permissions import OAuth2TokenPermission


class TokenView(oauth_views.TokenView):
    # There is a big flow of logic that happens around this (create_token_response) behind the scenes.
    #
    # oauth2_provider.views.TokenView inherits from oauth2_provider.views.mixins.OAuthLibMixin
    # That's where this method comes from originally.
    # Then *that* method ends up calling oauth2_provider.oauth2_backends.OAuthLibCore.create_token_response
    # Then *that* method ends up (ultimately) calling oauthlib.oauth2.rfc6749....
    def create_token_response(self, request):
        # Django OAuth2 Toolkit has a bug whereby refresh tokens are *never*
        # properly expired (ugh):
        #
        # https://github.com/jazzband/django-oauth-toolkit/issues/746
        #
        # This code detects and auto-expires them on refresh grant
        # requests.
        if request.POST.get('grant_type') == 'refresh_token' and 'refresh_token' in request.POST:
            hashed_refresh_token = hash_string(request.POST['refresh_token'], hasher=hashlib.sha256, algo="sha256")
            refresh_token = OAuth2RefreshToken.objects.filter(token=hashed_refresh_token).first()
            if refresh_token:
                expire_seconds = get_setting('OAUTH2_PROVIDER', {}).get('REFRESH_TOKEN_EXPIRE_SECONDS', 0)
                if refresh_token.created + timedelta(seconds=expire_seconds) < now():
                    return request.build_absolute_uri(), {}, 'The refresh token has expired.', '403'

        core = self.get_oauthlib_core()  # oauth2_provider.views.mixins.OAuthLibMixin.create_token_response

        # oauth2_provider.oauth2_backends.OAuthLibCore.create_token_response
        # (we override this so we can implement our own error handling to be compatible with AWX)

        # This is really, really ugly. Modify the request to hash the refresh_token
        # but only long enough for the oauth lib to do its magic.
        did_hash_refresh_token = False
        old_post = request.POST
        if 'refresh_token' in request.POST:
            did_hash_refresh_token = True
            request.POST = request.POST.copy()  # so it's mutable
            hashed_refresh_token = hash_string(request.POST['refresh_token'], hasher=hashlib.sha256, algo="sha256")
            request.POST['refresh_token'] = hashed_refresh_token

        try:
            uri, http_method, body, headers = core._extract_params(request)
        finally:
            if did_hash_refresh_token:
                request.POST = old_post

        extra_credentials = core._get_extra_credentials(request)
        try:
            headers, body, status = core.server.create_token_response(uri, http_method, body, headers, extra_credentials)
            uri = headers.get("Location", None)
            status = 201 if request.method == 'POST' and status == 200 else status
            return uri, headers, body, status
        except oauth2.AccessDeniedError as e:
            return request.build_absolute_uri(), {}, str(e), 403  # Compat with AWX
        except oauth2.OAuth2Error as e:
            return request.build_absolute_uri(), {}, str(e), e.status_code


class OAuth2TokenViewSet(ModelViewSet, AnsibleBaseDjangoAppApiView):
    queryset = OAuth2AccessToken.objects.all()
    serializer_class = OAuth2TokenSerializer
    permission_classes = [OAuth2ScopePermission, OAuth2TokenPermission]
